""" Bid for Game
    Responder's Bid class
"""
import inspect

from .bidding import Bid, Pass, Double, HandSuit
from .hand import Hand
from bridgeobjects import Suit

inspection = inspect.currentframe


class RespondersBid(Hand):
    """Responder's bid class."""

    def __init__(self, hand_cards, board):
        super(RespondersBid, self).__init__(hand_cards, board)
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        if self.overcaller_bid_one.level >= 4:
            bid = Pass('3001')
        elif self.opener_bid_one.is_nt:
            bid = self._respond_to_nt()
        elif self.opener_bid_one.level == 1:
            bid = self._respond_to_one_suit()
        elif self.opener_bid_one.name == '2C':
            bid = self._respond_to_two_clubs()
        elif self.opener_bid_one.level == 2:
            bid = self._respond_to_weak_two()
        elif self.opener_bid_one.level == 3:
            bid = self._respond_to_weak_three()
        elif self.opener_bid_one.is_game:
            bid = Pass('3167')
        else:
            bid = Pass('3173')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_nt(self):
        """Return responses to 1NT or 2NT."""
        if self.opener_bid_one.level == 1:
            bid = self._respond_to_one_nt()
        elif self.opener_bid_one.level == 2:
            bid = self._respond_to_two_nt()
        elif self.opener_bid_one.level == 3:
            bid = self._respond_to_three_nt()
        else:
            bid = Pass('3201')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_one_nt(self):
        """Respond to 1NT."""
        if self.overcaller_bid_one.is_value_call:
            bid = self._one_nt_with_overcall()
        elif self.hcp <= 10:
            bid = self._weak_take_out()
        else:
            bid = self._one_nt_test_invite()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _one_nt_with_overcall(self):
        """Return bid after 1NT and overcall."""
        if self._six_card_major_thirteen_points():
            bid = self.suit_bid(4, self.longest_suit, '3004')
        elif self._can_bid_after_one_nt_with_overcall():
            bid = self._trial_bid_suit_after_one_nt_with_overcall()
        elif self._is_semi_balanced_eleven_plus():
            bid = self._try_nt_after_one_nt_with_overcall()
        elif self._is_balanced_game_going() and self.five_card_major:
            bid = self.suit_bid(3, self.longest_suit, '3005')
        elif self._is_balanced_game_going():
            bid = self.nt_bid(3, '3006')
        elif self._is_game_going():
            bid = self.next_level_bid(self.longest_suit, '3124')
        elif (self.shape[0] >= 6 and
              self.hcp >= 10 and
              self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '3008')
        else:
            bid = Pass('3009')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _try_nt_after_one_nt_with_overcall(self):
        """Try a NT after 1NT and overcall."""
        if self.hcp >= 13:
            level = 3
        else:
            level = 2
        if level >= self.nt_level:
            bid = self.nt_bid(level, '3010')

        elif self.overcaller_bid_one.level >= 3:
            bid = Pass('3011')
        else:
            bid = Pass('3181')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_take_out(self):
        """Return bid for weak-takeout."""
        if self.shape[0] >= 5:
            bid = self._weak_take_out_bid()
        else:
            bid = Pass('3012')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_take_out_bid(self):
        """Return bid for weak-takeout."""
        suit = self.long_suit(5)
        if self.overcaller_bid_one.is_double:
            bid = self.suit_bid(2, suit, '3013')
        elif (self.shape[0] >= 6 and self.shape[1] >= 5 and self.hcp >= 10
              and self.longest_suit.is_major):
            bid = self.suit_bid(4, self.longest_suit, '3014')
        elif suit != self.club_suit:
            bid = self.suit_bid(2, suit, '3015')
        else:
            bid = Pass('3016')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _one_nt_test_invite(self):
        """Bid with four card major."""
        if 11 <= self.hcp <= 12:
            bid = self._one_nt_test_invite_weak()
        elif self.hcp >= 13:
            bid = self._respond_to_one_nt_strong()
        else:
            bid = Pass('3168')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _one_nt_test_invite_weak(self):
        if self.shape[0] >= 6:
            bid = self._respond_to_one_nt_strong()
        elif self.four_card_major:
            bid = self._one_nt_weak_stayman()
        elif self.nt_level <= 2:
            bid = self.nt_bid(2, '3018')
        else:
            bid = Pass('3169')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _one_nt_weak_stayman(self):
        if self.overcaller_bid_one.is_double and self.hcp >= 10:
            bid = Pass('3021')
        elif self.overcall_made:
            bid = Pass('3023')
        else:
            bid = self.club_bid(2, '3020')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_one_nt_strong(self):
        """Bid after 1NT strong."""
        if self._four_card_major_after_one_nt():
            bid = self._four_card_major_after_one_nt()
        elif self.shape[0] >= 6:
            bid = self.suit_bid(3, self.longest_suit, '3024')
        else:
            bid = self._strong_nt()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _four_card_major_after_one_nt(self):
        """Bid after 1NT, four card major."""
        if self.spades >= 6:
            bid = self.spade_bid(4, '3025')
        elif self.hearts >= 6:
            bid = self.heart_bid(4, '3026')
        elif self.spades >= 5:
            bid = self.spade_bid(3, '3027')
        elif self.hearts >= 5:
            bid = self.heart_bid(3, '3028')
        elif (not self.overcall_made and
              self.hcp <= 18 and
              self.four_card_major):
            bid = self.club_bid(2, '3029')  # Stayman
        else:
            bid = None
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _strong_nt(self):
        """Bid after 1NT, no four card major."""
        if self.hcp <= 18:
            bid = self.nt_bid(3, '3030')
        elif 19 <= self.hcp <= 20:
            bid = self.nt_bid(4, '3031')
        elif 21 <= self.hcp <= 22:
            bid = self.nt_bid(6, '3032')
        elif 23 <= self.hcp <= 24:
            bid = self.nt_bid(5, '3033')
        else:  # elif self.hcp >= 25:
            bid = self.nt_bid(7, '3034')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_two_nt(self):
        """Responses to 2NT."""
        if self.hcp <= 3:
            bid = Pass('3035')
        elif (self.spades >= 5 and
              self.hearts >= 5 and
              self.shape[0] == self.shape[1] and
              self.bid_history[-1] == 'P'):
            bid = self.next_level_bid(self.spade_suit, '3036')
        elif (self.spades >= 4 and
              self.hearts >= 4 and
              self.shape[1] <= 4 and
              self.bid_history[-1] == 'P' and
              self.next_level(self.club_suit) <= 3):
            bid = self.club_bid(3, '3037')
        elif self.spades >= 6 and self.next_level(self.spade_suit) <= 4:
            bid = self.spade_bid(4, '3041')
        elif self.hearts >= 6 and self.next_level(self.heart_suit) <= 4:
            bid = self.heart_bid(4, '3041')
        elif self.spades >= 5 and self.next_level(self.spade_suit) <= 3:
            bid = self.spade_bid(3, '3041')
        elif self.hearts >= 5 and self.next_level(self.heart_suit) <= 3:
            bid = self.heart_bid(3, '3041')
        elif ((self.spades >= 4 or self.hearts >= 4) and
              self.next_level(self.club_suit) <= 3):
            bid = self.club_bid(3, '3042')
        elif 11 <= self.hcp <= 12 and self.nt_level <= 4:
            bid = self.nt_bid(4, '3043')
        elif 13 <= self.hcp <= 14 and self.nt_level <= 6:
            bid = self.nt_bid(6, '3044')
        elif 15 <= self.hcp <= 16 and self.nt_level <= 5:
            bid = self.nt_bid(5, '3045')
        elif self.hcp >= 17 and self.nt_level <= 7:
            bid = self.nt_bid(7, '3046')
        elif self.nt_level <= 3:  # self.hcp >= 4:
            bid = self.nt_bid(3, '3047')
        elif self.double_allowed():
            bid = Double('3171')
        else:
            bid = Pass('3204')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_one_suit(self):
        """Respond to one of suit."""
        if self.opener_suit_one.is_minor:
            bid = self._minor_opening_check_major()
        else:
            bid = self._minor_opening_check_support()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _minor_opening_check_major(self):
        """Respond after one of minor with four card major."""
        if self.four_card_major:
            bid = self._respond_to_one_suit_no_support()
            if bid.is_pass:
                bid = self._minor_opening_check_support()
        else:
            bid = self._respond_to_minor()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_three_nt(self):
        """Responses to 2NT."""
        if self.hcp <= 8:
            bid = Pass('3003')
        # elif (self.spades >= 5 and
        #       self.hearts >= 5 and
        #       self.shape[0] == self.shape[1] and
        #       self.bid_history[-1] == 'P'):
        #     bid = self.next_level_bid(self.spade_suit, '3036')
        # elif (self.spades >= 4 and
        #       self.hearts >= 4 and
        #       self.shape[1] <= 4 and
        #       self.bid_history[-1] == 'P' and
        #       self.next_level(self.club_suit) <= 3):
        #     bid = self.club_bid(3, '3037')
        # elif self.spades >= 6 and self.next_level(self.spade_suit) <= 4:
        #     bid = self.spade_bid(4, '3041')
        # elif self.hearts >= 6 and self.next_level(self.heart_suit) <= 4:
        #     bid = self.heart_bid(4, '3041')
        # elif self.spades >= 5 and self.next_level(self.spade_suit) <= 3:
        #     bid = self.spade_bid(3, '3041')
        # elif self.hearts >= 5 and self.next_level(self.heart_suit) <= 3:
        #     bid = self.heart_bid(3, '3041')
        # elif ((self.spades >= 4 or self.hearts >= 4) and
        #       self.next_level(self.club_suit) <= 3):
        #     bid = self.club_bid(3, '3042')
        # elif 11 <= self.hcp <= 12 and self.nt_level <= 4:
        #     bid = self.nt_bid(4, '3043')
        # elif 13 <= self.hcp <= 14 and self.nt_level <= 6:
        #     bid = self.nt_bid(6, '3044')
        # elif 15 <= self.hcp <= 16 and self.nt_level <= 5:
        #     bid = self.nt_bid(5, '3045')
        # elif self.hcp >= 17 and self.nt_level <= 7:
        #     bid = self.nt_bid(7, '3046')
        # elif self.nt_level <= 3:  # self.hcp >= 4:
        #     bid = self.nt_bid(3, '3047')
        else:
            bid = Pass('3182')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _minor_opening_check_support(self):
        """Respond after one of minor, check support."""
        if self.suit_length(self.opener_suit_one) >= 4:
            bid = self._respond_to_one_suit_with_support()
        else:
            bid = self._respond_to_one_suit_no_support()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_minor(self):
        """Return bid after opener has bid minor suit."""
        # Absence of a 4 card major already accounted for
        hand_value_points = self.support_points(self.opener_suit_one)
        if hand_value_points <= 9:
            minor_level = 2
        elif hand_value_points <= 10:
            minor_level = 3
        elif hand_value_points <= 13:
            minor_level = 4
        else:
            minor_level = 5
        if hand_value_points <= 4:
            bid = Pass('3049')
        elif (self.hcp >= 16 and
              self.second_suit == self.opener_suit_one and
              self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '3050', raise_level=1)
        # elif self.hcp >=14 and self.suit_length(self.opener_suit_one) >=4:
        #     bid = self.next_level_bid(self.second_suit, '3167', raise_level=1)
        elif (self._is_five_four_game_going() and
              self.second_suit == self.opener_suit_one):
            bid = self.next_level_bid(self.second_suit, '3051', raise_level=1)
        elif self._is_balanced_with_no_long_minor():
            bid = self._nt_response_to_minor()
        elif self._has_six_card_unbid_suit():
            if self.hcp >= 16:
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(self.longest_suit, '3052', raise_level)
        elif self._is_semi_balanced_some_support():
            bid = self.nt_bid(2, '3053')
        elif self.opener_suit_one == self.club_suit and minor_level > 0:
            bid = self._respond_to_clubs(minor_level)
        elif self.opener_suit_one == self.diamond_suit and minor_level > 0:
            bid = self._response_to_diamonds(minor_level)
        else:
            bid = self._respond_to_one_suit_no_support()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _nt_response_to_minor(self):
        """Bid NT in response to minor opening."""
        stopper = self.stoppers_in_bid_suits
        if self.hcp >= 16 and self.longest_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.longest_suit, '3054', raise_level=1)
        elif self.hcp >= 13 and stopper and self.nt_level <= 3:
            bid = self.nt_bid(3, '3055')
        elif self.hcp >= 10 and stopper and self.nt_level <= 2:
            bid = self.nt_bid(2, '3056')
        elif self.hcp >= 6 and stopper and self.nt_level <= 1:
            bid = self.nt_bid(1, '3057')
        elif self.hcp <= 10 and self.nt_level > 1:
            bid = Pass('3058')
        elif self.spades <4 and self.hearts < 4:
            bid = Pass('3178')
        else:
            bid = Pass('3190')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_clubs(self, level):
        """Respond after club opening."""
        club_level = self.next_level(self.club_suit)
        hand_value_points = self.support_points(self.opener_suit_one)
        if (level > 3 and self.hcp >= 13 and self.nt_level <= 3 and
                self.shape[2] >= 3 and
                self.stoppers_in_bid_suits):
            bid = self.nt_bid(3, '3059')
        elif self._has_five_card_club_support_and_ten_points():
            if self.hcp >= 15:
                level = 5
            elif self.hcp >= 12:
                level = 4
            else:
                level = 3
            bid = self.suit_bid(level, self.club_suit, '3060')
        elif self._has_five_card_suit_and_ten_points():
            bid = self.next_level_bid(self.longest_suit, '3061')
        elif (self.clubs >= 4 and
              hand_value_points >= 5 and
              club_level <= level):
            bid = self.club_bid(level, '3062', True)
        elif self.diamonds >= 5:
            bid = self._respond_to_clubs_with_diamonds()
        elif self.clubs >= 4:
            bid = self._respond_to_one_suit_with_support()
        else:
            bid = self._respond_to_one_suit_no_support()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_clubs_with_diamonds(self):
        """Respond after club opening with diamonds."""
        if self.diamond_suit in self.opponents_suits:
            bid = Pass('3063')
        elif self.hcp + self.shape[0] >= 16:
            bid = self.next_level_bid(self.diamond_suit, '3064')
        elif self._can_rebid_diamonds_at_next_level():
            bid = self.next_level_bid(self.diamond_suit, '3065')
        elif self.stoppers_in_bid_suits and 6 <= self.hcp <= 9 and self.nt_level == 1:
            bid = self.nt_bid(1, '3191')
        elif not self.stoppers_in_bid_suits:
            bid = Pass('3066')
        elif self.hcp <= 8 and self.nt_level > 1:
            bid = Pass('3179')
        else:
            bid = Pass('3192')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _response_to_diamonds(self, level):
        """Respond after diamond opening."""
        club_level = self.next_level(self.club_suit)
        diamond_level = self.next_level(self.diamond_suit)
        hand_value_points = self.support_points(self.opener_suit_one)
        if (6 <= self.hcp <= 9 and
                self.is_balanced and self.stoppers_in_bid_suits and
                self.nt_level == 1):
            bid = self.nt_bid(1, '3067')
        elif self.is_balanced and 10 <= hand_value_points <= 12:
            bid = self.diamond_bid(3, '3068', True)
        elif self.diamonds >= 4 and 10 <= hand_value_points <= 14:
            level = max(level, diamond_level)
            bid = self.diamond_bid(level, '3069', True)
        elif self.clubs >= 5 and hand_value_points >= 9 and club_level <= 2:
            level = 2
            if self.hcp >= 16:
                level = 3
            bid = self.club_bid(level, '3070', True)
        elif self.diamonds >= 4:
            bid = self._respond_to_one_suit_with_support()
        else:
            bid = self._respond_to_one_suit_no_support()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_one_suit_with_support(self):
        """Respond with support for opener."""
        suit = self.opener_suit_one
        level = self.next_level(suit)
        hand_value_points = (self.hcp
                             + self.support_shape_points(suit)
                             + self.suit_length(suit) - 4)

        if suit.is_minor:
            bid = self._respond_to_one_minor_with_support()
        elif hand_value_points < 6:
            bid = Pass('3080', True)
        elif (6 <= hand_value_points <= 9 and
              level <= self.opener_bid_one.level + 1):
            bid = self.suit_bid(level, suit, '3072', True)
        elif (10 <= hand_value_points <= 11 and
              level <= self.opener_bid_one.level + 2):
            bid = self.suit_bid(level + 1, suit, '3202', True)
        elif 12 <= hand_value_points <= 15 and level <= self.opener_bid_one.level + 3:
            if self.suit_length(suit) >= 5 and self.hcp >= 11:
                level += (self.suit_length(suit) - 4)
            if 10 <= self.hcp <= 11:
                level -= 1
            level = min(2, level)
            bid = self.suit_bid(level + 2, suit, '3203', True)
        elif suit.is_major and hand_value_points >= 18:
            bid = self.nt_bid(4, '3084', True)  # Blackwood
        elif self.hcp >= 16 and suit.is_minor:
            bid = self.suit_bid(6, suit, '3085')
        elif (self.hcp >= 16 and
              self.shape[1] >= 4 and
              self.longest_suit not in self.opponents_suits):
            if suit == self.longest_suit and self.second_suit not in self.opponents_suits:
                suit = self.second_suit
            else:
                suit = self.longest_suit
            bid = self.next_level_bid(suit, '3086', raise_level=1)
        elif hand_value_points >= 13:
            bid = self.bid_to_game(suit, '3087')
        else:
            bid = Pass('3088', True)
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_one_minor_with_support(self):
        """Respond with support for opener."""
        suit = self.opener_suit_one
        level = self.next_level(suit)
        hand_value_points = (self.hcp
                             + self.support_shape_points(suit)
                             + self.suit_length(suit) - 4)
        if hand_value_points < 6:
            bid = Pass('3080', True)
        elif (6 <= hand_value_points <= 9 and
              level <= self.opener_bid_one.level + 1):
            bid = self.suit_bid(level, suit, '3081', True)
        elif (10 <= hand_value_points <= 11 and
              level <= self.opener_bid_one.level + 2):
            bid = self.suit_bid(level + 1, suit, '3105', True)
        elif 12 <= hand_value_points <= 15 and level <= self.opener_bid_one.level + 3:
            if self.suit_length(suit) >= 5 and self.hcp >= 11:
                level += (self.suit_length(suit) - 4)
            if 10 <= self.hcp <= 11:
                level -= 1
            level = min(2, level)
            bid = self.suit_bid(level + 2, suit, '3083', True)
        elif suit.is_major and hand_value_points >= 18:
            bid = self.nt_bid(4, '3084', True)  # Blackwood
        elif self.hcp >= 16 and suit.is_minor:
            bid = self.suit_bid(6, suit, '3085')
        elif (self.hcp >= 16 and
              self.shape[1] >= 4 and
              self.longest_suit not in self.opponents_suits):
            if suit == self.longest_suit and self.second_suit not in self.opponents_suits:
                suit = self.second_suit
            else:
                suit = self.longest_suit
            bid = self.next_level_bid(suit, '3086', raise_level=1)
        elif hand_value_points >= 13:
            bid = self.bid_to_game(suit, '3087')
        else:
            bid = Pass('3088', True)
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_one_suit_no_support(self):
        """Respond to opening suit bid with no support."""
        own_suit = self._suit_for_response_no_support()
        response_level = self._level_for_response(own_suit)
        if own_suit == self.no_trumps:
            bid = self._no_support_bid_nt()
        elif self.hcp >= 16:
            bid = self._no_support_strong(own_suit, response_level)
        elif 9 <= self.hcp <= 15:
            bid = self._no_support_intermediate(response_level, own_suit, '3099')
        elif 6 <= self.hcp <= 8:
            bid = self._no_support_weak(response_level, own_suit)
        elif self.hcp == 5 and response_level == 1 and self.shape[0] >= 5:
            bid = self.suit_bid(1, own_suit, '3090')
        else:  # elif self.hcp <= 5:
            bid = Pass('3091')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_bid_nt(self):
        """Respond to opening suit bid with no support, nt."""
        if self.hcp >= 8:
            stoppers = self.four_in_bid_suits()
        else:
            stoppers = self.stoppers_in_bid_suits
        poor_stoppers = self.poor_stoppers_in_bid_suits
        hold_opponents_suit = False
        if (self.overcaller_bid_one.is_suit_call and
                self.suit_length(self.overcaller_bid_one.denomination) >= 5):
            hold_opponents_suit = True

        if self.hcp <= 9 and hold_opponents_suit:
            bid = Pass('3092')
        elif stoppers or (self.shape[0] == 4 and poor_stoppers):
            bid = self._bid_nt_stoppers()
        elif self.hcp <= 9:
            bid = Pass('3172')
        else:
            bid = Pass('3187')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _bid_nt_stoppers(self):
        """Return NT bid with stoppers."""
        if self.nt_level <= 3 and self.hcp >= 13:
            bid = self.nt_bid(3, '3094')
        elif self.nt_level <= 2 and self.hcp >= 10:
            bid = self.nt_bid(2, '3095')
        elif self.nt_level == 1 and self.hcp >= 6:
            bid = self.nt_bid(1, '3096')
        else:
            bid = Pass('3097')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_strong(self, own_suit, response_level):
        """Respond to opening suit bid with no support, strong."""
        level = self._get_level__no_support_strong(own_suit, response_level)
        bid = self.suit_bid(level, own_suit, '3098')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_intermediate(self, response_level, own_suit, comment=''):
        """Return bid with no support for opener's suit, intermediate."""
        comment = '3106'
        if own_suit == self.opener_suit_one:
            comment = '3105'
        trial_bid = self.suit_bid(response_level, own_suit, comment)

        if (self.opener_bid_one.name == '1S' and
                trial_bid.name == '2H' and self.hearts < 5):
            bid = self._no_support_intermediate_after_spade()
        elif self._has_five_card_major_and_game_going():
            bid = self.next_level_bid(self.five_card_major_suit, '3099')
        elif self._is_balanced_and_strong():
            bid = self.nt_bid(3, '3132')
        elif self._is_game_going_support_for_partners_minor(response_level) and self.double_allowed():
            bid = Double('3101')
        elif (self._has_four_card_major_and_game_going(response_level) and
              self.longest_suit == self.opener_suit_one):
            suit = self.next_four_card_suit()
            bid = self.next_level_bid(suit, '3102')
        elif self._has_overcall_but_no_support(own_suit):
            if self.hcp >= 14:
                level = 3
            else:
                level = 2
            bid = self.nt_bid(level, '3103')
        else:
            bid = self._no_support_intermediate_not_spade(own_suit, trial_bid)
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_intermediate_after_spade(self):
        """Return bid with no support for opener's spade."""
        if self._has_good_diamonds():
            bid = self.diamond_bid(2, '3107')
        elif self._has_good_clubs():
            bid = self.club_bid(2, '3108')
        # else:
        hcp = self.hcp
        level = self.quantitative_raise(hcp, 0, [0, 10, 13], 3)
        if level >= self.nt_level:
            bid = self.nt_bid(level, '3109')
        elif self.hcp <= 9 and self.nt_level > 1:
            bid = Pass('3110')
        else:
            bid = Pass('3180')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_intermediate_not_spade(self, own_suit, trial_bid):
        """Return bid with no support for opener's suit, intermediate."""
        if own_suit.is_minor and self.is_balanced and self.hcp == 9:
            bid = self._no_support_intermediate_minor_balanced()
        elif self._is_semi_balanced_no_support(own_suit) and self.hcp <= 12:
            bid = self._no_support_intermediate_minor_balanced_overcall()
        elif self._four_four_one_stoppers():
            bid = self._singleton_in_openers_suit(trial_bid)
        elif self._has_singleton_in_openers_suit_and_stoppers(own_suit):
            bid = self.nt_bid(2, '3112')
        elif self._has_no_support_and_no_long_suit(own_suit, trial_bid):
            level = 2
            comment = '3114'
            if self.hcp >= 13:
                level = 3
                comment = '3165'
            bid = self.nt_bid(level, comment)
        elif self._forced_to_jump_bid(own_suit, trial_bid):
            bid = Pass('3115')
        else:
            bid = trial_bid
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _singleton_in_openers_suit(self, trial_bid):
        """Return bid if holding a singleton in openers suit."""
        if self.overcaller_bid_one.is_nt and self.nt_level == 2:
            if self.hcp >= 10:
                bid = self.nt_bid(2, '3116')
            else:
                bid = Pass('3117')
        # else:
        #     bid = self.nt_bid(2, '3118')
        else:
            bid = trial_bid
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_intermediate_minor_balanced(self):
        """Return bid with no support balanced but minor suit."""
        if self.nt_level == 1 and self.stoppers_in_bid_suits:
            if self.shape[1] == 4 and self.next_level(self.longest_suit) == 1:
                bid = self.next_level_bid(self.longest_suit, '3119')
            else:
                bid = self.nt_bid(1, '3120')
        elif self.nt_level >= 1 or not self.stoppers_in_bid_suits:
            bid = Pass('3121')
        else:
            bid = Pass('3185')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_intermediate_minor_balanced_overcall(self):
        """Return bid with no support balanced, overcall, minor suit."""
        if self.overcaller_bid_one.name == '1NT' and self.hcp >= 10 and self.double_allowed():
            bid = Double('3122')
        else:
            if self.hcp >= 13:
                level = 3
            elif self.hcp >= 10:
                level = 2
            else:
                level = self.nt_level
            if level >= self.nt_level:
                bid = self.nt_bid(level, '3123')
            else:
                bid = Pass('3124')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_weak(self, response_level, own_suit):
        """Respond to opening suit bid with no support, weak."""
        if response_level == 1 and self.next_level(own_suit) == 1:
            bid = self.suit_bid(1, own_suit, '3125')
            # bid = self._no_support_weak_level_one(response_level, own_suit)
        else:
            bid = self._no_support_weak_over_level_one(own_suit)
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    # def _no_support_weak_level_one(self, level, own_suit):
    #     """Respond to bid with no support at level one, weak."""
    #     if self.next_level(own_suit) == 1:
    #         bid = self.suit_bid(level, own_suit, '3125')
    #     else:
    #         suit = self.second_suit
    #         level = self.next_level(suit)
    #         bid = self.suit_bid(level, suit, '3126')
    #     self.tracer(__name__, inspection(), bid, self.trace)
    #     return bid

    def _no_support_weak_over_level_one(self, own_suit):
        """Respond to bid with no support above level one, weak."""
        major = self.four_card_major_suit
        if major and self.next_level(major) == 1:
            bid = self.next_level_bid(major, '3127')
        elif (not self.overcaller_bid_one.is_pass and
              not self.overcaller_bid_one.is_double):
            bid = self._no_support_weak_over_level_one_overcall()
        # else:
        elif own_suit.is_minor:
            bid = self._no_support_weak_over_level_one_minor()
        elif self.stoppers_in_bid_suits:
            bid = self.nt_bid(1, '3128')
        else:
            bid = self.nt_bid(1, '3186')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_weak_over_level_one_overcall(self):
        """Respond to bid with no support above level one, overcall."""
        stopper = False
        if not self.overcaller_bid_one.is_nt:
            stopper = self.has_stopper(self.overcaller_bid_one.denomination)
        if stopper and self.nt_level <= 1:
            bid = self.nt_bid(1, '3129')
        elif (self.shape[0] >= 7 and
              self.hcp >= 8 and
              self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '3130')
        elif not self.overcaller_bid_one.is_nt and not stopper:
            bid = Pass('3131')
        elif self.hcp <= 9:
            bid = Pass('3177')
        else:
            bid = Pass('3199')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_weak_over_level_one_minor(self):
        """Respond to bid with no support above level one, minor."""
        if (self.hearts >= 4 and
                self.next_level(self.heart_suit) == 1):
            bid = self.heart_bid(1, '3132')
        elif (self.spades >= 4 and
              self.next_level(self.spade_suit) == 1):
            bid = self.spade_bid(1, '3133')
        elif self.shape[0] >= 8 and self.hcp >= 8 and self.next_level(self.longest_suit) <= 2:
            bid = self.next_level_bid(self.longest_suit, '3134')
        else:
            bid = self.nt_bid(1, '3135')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_two_clubs(self):
        """Rebid after 2C opening."""
        suit = self._get_suit_after_two_clubs_opening()
        quality = HandSuit(suit, self).suit_quality()
        suit_length = self.suit_length(self.longest_suit)
        if ((self.hcp > 7 or
             (self.hcp == 7 and
              quality >= 2)) and suit_length >= 5):
            bid = self._suit_response_after_two_clubs(suit)
        elif self.is_balanced and self.hcp >= 8:
            bid = self.nt_bid(2, '3136')
        elif (self.diamond_suit not in self.opponents_suits and
              self.next_level(self.diamond_suit) <= 2):
            bid = self.diamond_bid(2, '3137')
        elif self.diamond_suit in self.opponents_suits:
            bid = Pass('3138')
        elif (self.overcaller_bid_one.is_value_call and
              (self.shape[0] < 5 or self.hcp <= 2)):
            bid = Pass('3193')
        elif self.overcall_made:
            bid = Pass('3200')
        else:
            bid = Pass('3184')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _suit_response_after_two_clubs(self, own_suit):
        """Make suit response after 2C."""
        if own_suit.is_major:
            bid = self.next_level_bid(own_suit, '3139')
        # if own_suit == self.spade_suit:
        #     bid = self.spade_bid(2, '3139')
        # elif own_suit == self.heart_suit:
        #     bid = self.heart_bid(2, '3140')
        elif own_suit == self.diamond_suit:
            level = self.next_level(self.diamond_suit)
            if level == 2:
                level = 3
            bid = self.diamond_bid(level, '3141')
        elif own_suit == self.club_suit:
            bid = self.next_level_bid(self.club_suit, '3142')
        else:
            bid = Pass('3175')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_weak_two(self):
        """Respond to weak two."""
        openers_suit = self.opener_suit_one
        hand_value_points = (self.hcp +
                             self.support_shape_points(openers_suit))
        if self.hcp <= 10:
            bid = self._respond_to_weak_two_weak()
        elif self.overcaller_bid_one.is_double:
            bid = Pass('3143')
        elif self._support_after_suit_overcall_game_going():
            level = self.next_level(openers_suit)
            bid = self.suit_bid(level, openers_suit, '3144', True)
        elif hand_value_points >= 16:
            bid = self._respond_to_weak_two_strong()
        elif self._has_singleton_in_openers_suit_and_own_six_card_suit():
            bid = self.next_level_bid(self.longest_suit, '3145')
        elif self.hcp <= 15:
            bid = Pass('3146')
        else:
            bid = Pass('3174')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_weak_two_weak(self):
        """Respond to weak two if weak."""
        openers_suit = self.opener_suit_one
        level = self.next_level(openers_suit)
        if self.suit_length(openers_suit) >= 4 and level <= 4:
            bid = self.suit_bid(4, openers_suit, '3147')
        elif self.suit_length(openers_suit) >= 3 and level <= 3:
            bid = self.suit_bid(3, openers_suit, '3148')
        else:
            bid = Pass('3149')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_weak_two_strong(self):
        """Respond to weak two if strong."""
        if self._has_two_card_support_for_major():
            bid = self.suit_bid(4, self.opener_suit_one, '3150', True)
        elif self.hcp >= 16 and self._shape[0] >= 8 and self.longest_suit.is_major:
            bid = self.bid_to_game(self.longest_suit, '3151')
        elif self._has_own_suit():
            # fewer than 2 in partner's suit checked in _has_own_suit
            bid = self.next_level_bid(self.longest_suit, '3152')
        elif (self.hcp >= 23 and
              self.suit_length(self.opener_suit_one) >= 2 and
              self.is_balanced):
            bid = self.nt_bid(3, '3153')
        elif self.hcp >= 23 and self.suit_length(self.opener_suit_one) >= 2:
            bid = self.suit_bid(6, self.opener_suit_one, '3154')
        elif (self.hcp >= 15 and
                self.shape[1] >= 5 and
                self.longest_suit.is_major and
                self.second_suit.is_major and
                self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '3155')
        else:
            if self.nt_level <= 3:
                bid = self.nt_bid(3, '3156')
            else:
                bid = Pass('3170')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_weak_three(self):
        """Respond to weak three."""
        if self.hcp >= 15:
            bid = self._respond_to_weak_three_strong()
        elif (self.hcp <= 9 and
              self.suit_length(self.opener_suit_one) >= 2):
            bid = self.suit_bid(4, self.opener_suit_one, '3158')
        elif self.hcp <= 14 and self.suit_length(self.opener_suit_one) <= 1:
            bid = Pass('3159')
        elif self.hcp <= 14:
            bid = Pass('3176')
        else:
            bid = Pass('3189')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_weak_three_strong(self):
        """Respond to weak three, strong."""
        if self._two_card_support_for_major_fifteen_points():
            bid = self.suit_bid(4, self.opener_suit_one, '3160')
        elif self.is_balanced and self.nt_level <= 3 and self.stoppers_in_bid_suits:
            bid = self.nt_bid(3, '3161')
        elif (self.opener_suit_one.is_major and
              self.suit_length(self.opener_suit_one) >= 1):
            bid = self.suit_bid(4, self.opener_suit_one, '3162')
        elif self._can_bid_six_card_suit_at_level_three():
            bid = self.next_level_bid(self.longest_suit, '3163')
        elif self.nt_level <= 3 and self.stoppers_in_bid_suits:
            bid = self.nt_bid(3, '3164')
        else:
            bid = Pass('3165')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    # Various utility functions

    def _get_suit_after_two_clubs_opening(self):
        """Determine best suit after 2C opening."""
        if self.five_five:
            if self.longest_suit < self.second_suit:
                suit = self.second_suit
            else:
                suit = self.longest_suit
        else:
            suit = self.longest_suit
        return suit

    def _get_one_nt_jump_level(self):
        """Return jump level for response after 1NT with overcall."""
        if self.hcp >= 12:
            jump_level = 1
        else:
            jump_level = 0
        return jump_level

    def _level_for_response(self, own_suit):
        """Return level for no support bid."""
        level = self.next_level(own_suit)
        if self._has_overcall_ten_points_shortage(own_suit):
            level += 1
        elif self._has_support_nine_points_shortage(own_suit):
            level += 1
        return level

    def _get_level__no_support_strong(self, own_suit, level):
        """Return level for strong bid with no support."""
        if self.longest_suit.is_major and self.shape[0] >= 6:
            level += 2
        else:
            level += 1
        jump_level = self.next_level(own_suit) + 1
        level = min(level, jump_level)
        return level

    def _suit_for_response_no_support(self):
        """Return suit for responder when no support for opener."""
        if self.five_five_or_better:
            next_suit = self._higher_five_card_suit()
        else:
            next_suit = self._cheapest_four_card_suit(self.opener_suit_one)
        suit = self.longest_suit
        if suit in self.opponents_suits:
            suit = self.second_suit
        if self.suit_length(suit) == 4:
            for bid in self.bid_history[::-1]:
                if Bid(bid).is_value_call:
                    suit = self._cheapest_four_card_suit(Bid(bid).denomination)
                    break
        if self.suit_length(suit) < 4:
            suit = self.no_trumps
        elif suit == self.overcaller_bid_one.denomination:
            suit = self.no_trumps
        elif (self.hcp >= 16 and
              self.longest_suit == self.opener_suit_one and
              self.shape[1] == 4 and
              next_suit and
              next_suit not in self.opponents_suits):
            suit = next_suit
        return suit

    def _cheapest_four_card_suit(self, bid_suit):
        """Return cheapest 4 card suit after given suit."""
        loop = 0
        rank = bid_suit.rank
        while True:
            rank = (rank + 1) % 4
            if self.suit_length(self.suits[rank]) >= 4:
                break
            loop += 1
            if loop > 4:
                rank = 4
                break
        suit_name = ['C', 'D', 'H', 'S', None][rank]
        suit = Suit(suit_name)
        return suit

    def _higher_five_card_suit(self):
        """Return highest five card suit."""
        for rank in list(range(4))[::-1]:
            if self.suit_length(self.suits[rank]) >= 5:
                suit_name = ['C', 'D', 'H', 'S'][rank]
                suit = Suit(suit_name)
                return suit

    def _get_best_major(self):
        """Return the best major suit."""
        if self.spades >= 6:
            suit = self.spade_suit
        elif self.hearts >= 6:
            suit = self.heart_suit
        elif self.spades == 5:
            suit = self.spade_suit
        elif self.hearts == 5:
            suit = self.heart_suit
        elif self.hearts == 4:
            suit = self.heart_suit
        else:
            suit = self.spade_suit
        return suit

    def _one_nt_max_level(self):
        """Return maximum level for response after 1NT with overcall."""
        if (self.hcp >= 12 or
                (self.hcp == 11 and self.shape[0] >= 6)):
            max_level = 3
        elif self.hcp >= 8:
            max_level = 2
        elif self.hcp >= 6 and self.shape[0] >= 6:
            max_level = 2
        else:
            max_level = 1
        return max_level

    def _trial_bid_suit_after_one_nt_with_overcall(self):
        """Try a suit after 1NT and overcall."""
        jump_level = self._get_one_nt_jump_level()
        trial_bid = self.next_level_bid(self.longest_suit, '3166', jump_level)
        return trial_bid

    # Various boolean functions

    def _six_card_major_thirteen_points(self):
        """Return True if hand has a six+ card_major with 13 points."""
        return (self.shape[0] >= 6 and
                self.longest_suit not in self.opponents_suits and
                self.longest_suit.is_major and
                self.hcp >= 13)

    def _can_bid_after_one_nt_with_overcall(self):
        """Return True if suit bid can be made after 1NT with overcall."""
        max_level = self._one_nt_max_level()
        trial_bid = self._trial_bid_suit_after_one_nt_with_overcall()
        return (self.shape[0] >= 5 and
                trial_bid.denomination not in self.opponents_suits and
                trial_bid.level <= max_level)

    def _is_semi_balanced_eleven_plus(self):
        """Return True if semi balanced and 11+ points and stoppers."""
        if self.is_semi_balanced and self.hcp >= 11:
            if self.stoppers_in_bid_suits:
                return True
            elif (self.suit_length(self.overcaller_bid_one.denomination) >= 2 and
                  self.suit_points(self.overcaller_bid_one.denomination) >= 2):
                return True
        return False

    def _is_balanced_game_going(self):
        """Return True if balanced with 13+ points."""
        return (self.hcp >= 13 and
                self.is_balanced and
                # self.stoppers_in_bid_suits and
                self.nt_level <= 3)

    def _is_game_going(self):
        """Return True with 13+ points."""
        return (self.hcp >= 13 and
                self.longest_suit not in self.opponents_suits)

    def _is_five_four_game_going(self):
        """Return True with 5/4 and 14+ points."""
        return (self.hcp >= 14 and
                self.five_four and
                self.second_suit not in self.opponents_suits)

    def _is_balanced_with_no_long_minor(self):
        """Return True if balanced with only 4 card minors."""
        return (self.is_balanced and
                self.clubs < 5 and self.diamonds < 5 and
                self.stoppers_in_bid_suits)

    def _has_six_card_unbid_suit(self):
        """Return True with 6+ card unbid suit."""
        return (self.shape[0] >= 6 and
                self.hcp >= 9 and
                self.opener_suit_one != self.longest_suit
                and self.longest_suit not in self.opponents_suits)

    def _is_semi_balanced_some_support(self):
        """Return True if semi balanced with 3+cards in partners suit."""
        return (self.is_semi_balanced and
                self.suit_length(self.opener_suit_one) >= 3 and
                self.stoppers_in_bid_suits and
                11 <= self.hcp <= 12 and
                self.longest_suit == self.opener_suit_one and
                self.nt_level <= 2 and
                self.shape[3] >= 2)

    def _has_five_card_club_support_and_ten_points(self):
        """Return True with 5+ support for clubs and 10+ points."""
        return (self.longest_suit == self.club_suit and
                self.shape[0] >= 5 and
                self.hcp >= 10 and
                self.longest_suit not in self.opponents_suits)

    def _has_five_card_suit_and_ten_points(self):
        """Return True with 5 card suits and 10+ points."""
        return (self.shape[0] >= 5 and
                10 <= self.hcp <= 15 and
                self.longest_suit not in self.opponents_suits)

    def _is_balanced_fewer_than_eight_points(self):
        """Return True if balanced and fewer than 8 points."""
        return (6 <= self.hcp <= 8 and
                self.is_balanced and
                self.stoppers_in_bid_suits and
                self.nt_level == 1)

    def _has_overcall_ten_points_shortage(self, own_suit):
        """Return True if overcall and 10 points and doubleton."""
        return (self.overcall_made and self.hcp >= 10 and
                self.shape[3] <= 2 and
                self.shape[self.opener_suit_one.rank] >= 2 and
                own_suit == self.opener_suit_one)

    def _has_support_nine_points_shortage(self, own_suit):
        """Return True if 5 card support and 9 points and singleton."""
        return (own_suit == self.opener_suit_one and
                self.suit_length(own_suit) >= 5 and
                self.shape[3] <= 1 and
                self.hcp >= 9)

    def _has_five_card_major_and_game_going(self):
        """Return True with 5 card major and 13+ points."""
        return (self.five_card_major_suit and
                self.hcp >= 13 and
                self.five_card_major_suit not in self.opponents_suits)

    def _is_balanced_and_strong(self):
        """Return True if balanced and 15+ points."""
        return (self.hcp >= 15 and
                self.shape[0] == 4 and
                self.stoppers_in_bid_suits and
                self.nt_level <= 3 and
                (not self.four_card_major_or_better or
                 self.overcall_made))

    def _has_four_card_major_and_game_going(self, response_level):
        """Return True with 4 card major and 13+ points."""
        return (self.four_card_major_suit and
                response_level >= 2 and
                self.hcp >= 13 and
                self.four_card_major_suit not in self.opponents_suits and
                (self.opener_suit_one != self.spade_suit or
                 self.four_card_major_suit != self.heart_suit)
                )

    def _has_overcall_but_no_support(self, suit):
        """Return True if no support with overcall."""
        return (self.nt_level == 2 and
                self.suit_length(suit) <= 4 and
                suit != self.opener_suit_one and
                self.overcaller_bid_one.is_suit_call
                and self.stoppers_in_bid_suits)

    def _has_good_diamonds(self):
        """Return True with good diamonds."""
        diamond_level = self.next_level(self.diamond_suit)
        return (diamond_level <= 2 and
                (self.diamonds >= 5 or (self.diamonds == 4 and self.hcp >= 10)))

    def _has_good_clubs(self):
        """Return True with good clubs."""
        club_level = self.next_level(self.club_suit)
        return (club_level <= 2 and
                (self.clubs >= 5 or (self.clubs == 4 and self.hcp >= 10)))

    def _is_semi_balanced_no_support(self, own_suit):
        """Return True if semi balanced and overcall."""
        return (self.stoppers_in_bid_suits and
                self.is_semi_balanced and
                own_suit.is_minor and
                self.overcall_made and
                (own_suit != self.opener_suit_one or
                 self.hcp >= 10))

    def _four_four_one_stoppers(self):
        """Return True if 4441 and singleton in opener's suit."""
        return (self.stoppers_in_bid_suits and
                self.shape[0] == 4 and
                self.suit_length(self.opener_suit_one) == 1)

    def _has_singleton_in_openers_suit_and_stoppers(self, own_suit):
        """Return True if singleton in opener's suit and stopper, level 2."""
        return (self.stoppers_in_bid_suits and
                self.overcall_made and
                self.suit_length(self.opener_suit_one) == 1
                and own_suit != self.longest_suit and
                self.hcp >= 10 and self.nt_level == 2)

    def _has_no_support_and_no_long_suit(self, own_suit, trial_bid):
        """Return True with no support and no long suit."""
        return (trial_bid.level >= 2 and
                self.suit_length(own_suit) < 5 and
                self.is_jump(self.opener_bid_one, trial_bid) and
                own_suit != self.opener_suit_one and
                self.is_balanced and
                self.hcp >= 10 and self.nt_level <= 2 and
                (self.stoppers_in_bid_suits or
                 (self.is_balanced and self.shape[3] >= 3)))

    def _forced_to_jump_bid(self, own_suit, trial_bid):
        """Return True if forced to jump."""
        return (trial_bid.level >= 2 and
                self.suit_length(own_suit) < 5 and
                self.is_jump(self.opener_bid_one, trial_bid) and
                self.hcp <= 10)

    def _support_after_suit_overcall_game_going(self):
        """Return True with support after overcall."""
        hand_value_points = (self.hcp +
                             self.support_shape_points(self.opener_suit_one))
        return (12 <= hand_value_points <= 15 and
                self.suit_length(self.opener_suit_one) >= 2 and
                self.overcaller_bid_one.is_value_call)

    def _has_singleton_in_openers_suit_and_own_six_card_suit(self):
        """Return True if singleton in opener's suit and a 6+card suit."""
        return (self.shape[0] >= 6 and
                self.hcp >= 12 and
                self.suit_length(self.opener_suit_one) <= 1 and
                self.longest_suit not in self.opponents_suits)

    def _has_two_card_support_for_major(self):
        """Return True if openers suit is major and two card support."""
        return (self.opener_suit_one.is_major and
                self.suit_length(self.opener_suit_one) >= 2)

    def _has_own_suit(self):
        if self.shape[0] >= 6:
            suit_quality = HandSuit(self.longest_suit, self).suit_quality()
            has_own_suit = True
        else:
            suit_quality = 0
            has_own_suit = False
        return (has_own_suit and suit_quality >= 2 and
                self.suit_length(self.opener_suit_one) <= 1 and
                self.longest_suit not in self.opponents_suits)

    def _two_card_support_for_major_fifteen_points(self):
        """Return True with support for opener's major."""
        return (self.opener_suit_one.is_major and
                self.suit_length(self.opener_suit_one) >= 2 and
                self.hcp >= 15 and
                self.next_level(self.opener_suit_one) <= 4)

    def _is_game_going_support_for_partners_minor(self, response_level):
        """Return True if game going and support for openers minor."""
        return (self._has_four_card_major_and_game_going(response_level) and
                self.next_four_card_suit() == self.opener_suit_one and
                self.opener_suit_one.is_minor)

    def _can_rebid_diamonds_at_next_level(self):
        """Return True if can bid diamonds at the next level."""
        hand_value_points = self.support_points(self.opener_suit_one)
        level = self.next_level(self.diamond_suit)
        return ((hand_value_points >= 5 and level <= 1) or
                (hand_value_points >= 9 and
                self.suit_points(self.diamond_suit) >= 5 and
                level <= 2))

    def _can_bid_six_card_suit_at_level_three(self):
        """Return True if Can bid six card suit at 3 level."""
        return (self.shape[0] >= 6 and
                self.next_level(self.longest_suit) <= 3 and
                self.longest_suit not in self.opponents_suits)



""" Bid for Game
    Acol AdvancersBid module
"""

import inspect

from bridgeobjects import Board, Card, Call, Suit

from bfgbidding.bidding import Bid, Pass, Double
from bfgbidding.hand import Hand

inspection = inspect.currentframe

# uncomment line around 30 for generating compliance tests. Affects hand 361!!!


class AdvancersBid(Hand):
    """BfG AdvancersBid class."""
    def __init__(self, hand_cards: list[Card], board: Board):
        super().__init__(hand_cards, board)
        self.overcallers_suit_holding = 0
        if self.overcaller_bid_one.is_suit_call:
            denomination = self.overcaller_bid_one.denomination
            self.overcallers_suit_holding = self.suit_holding[denomination]
        self.trace = 0

    def suggested_bid(self) -> Call:
        """Direct control to relevant method and return a Bid object."""
        if self._overcall_bid_at_three_level():
            bid = self._respond_to_three_level()
        elif self.overcaller_bid_one.is_double:
            bid = self._overcaller_has_doubled()
        elif (self.shape[0] >= 9 and
                # comment when not running immersion test
                # self.longest_suit not in self.opponents_suits and
                #
                self.next_level(self.longest_suit) <= 4):
            bid = self.suit_bid(4, self.longest_suit, '4001')
        elif self._has_biddable_eight_card_suit():
            bid = self.bid_to_game(self.longest_suit, '4002')
        elif self.overcaller_bid_one.is_nt:
            bid = self._respond_to_nt()
        elif self._has_three_card_support_for_overcallers_major():
            bid = self._support_for_overcaller()
        elif self._is_weak_no_support_for_overcaller():
            bid = self.nt_bid(1, '4003')
        elif self._has_biddable_suit():
            bid = self._long_suit_bid()
        elif self._can_bid_nt_opposition_has_no_fit():
            bid = self._no_support_stoppers()
        elif self._can_support_overcaller():
            bid = self._support_for_overcaller()
        elif self._has_opening_values_can_bid_nt():
            bid = self._bid_nt()
        elif self.hcp >= 12:
            bid = self._long_suit_bid()
        elif (self.overcaller_bid_one.is_suit_call
                and self.overcallers_suit_holding <= 3):
            bid = Pass('4125')
        else:
            bid = Pass('4004')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_three_level(self) -> Call:
        """Respond to 3 level bid."""
        jump_bid = self.is_jump(self.opener_bid_one, self.overcaller_bid_one)
        if self.overcaller_bid_one.is_game:
            bid = Pass('4005')
        elif self._can_bid_six_card_suit_at_three_level():
            bid = self.next_level_bid(self.longest_suit, '4006')
        elif self._has_ten_points_and_stoppers():
            bid = self.nt_bid(3, '4007')
        elif self._can_bid_to_game(self.overcaller_bid_one.denomination):
            bid = self.bid_to_game(
                self.overcaller_bid_one.denomination, '4008')
        elif self.hcp >= 10 and self.overcallers_suit_holding >= 4:
            bid = self.next_level_bid(
                self.overcaller_bid_one.denomination, '4009')
        elif (self.hcp <= 12 and self.opener_bid_one.level >= 2
                and not jump_bid):
            bid = Pass('4010')
        elif self._has_eight_points_and_stoppers():
            bid = self.nt_bid(3, '4011')
        elif ((self.overcallers_suit_holding <= 2 and
              self.hcp <= 8) or
              not self.stoppers_in_bid_suits):
            bid = Pass('4124')
        elif self.hcp <= 6:
            bid = Pass('4152')
        elif self.suit_holding[self.overcaller_suit_one] <= 3:
            bid = Pass('4154')
        elif self.overcaller_bid_one.level >= 4 and self.hcp <= 14:
            bid = Pass('4159')
        elif self.hcp <= 10 and self.next_level(self.overcaller_suit_one) >= 3:
            bid = Pass('4160')
        else:
            bid = Pass('4012')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _overcaller_has_doubled(self) -> Call:
        """Bid after partner has doubled."""
        if self._has_eleven_points_balanced_and_responder_has_bid_game():
            bid = Double('4013')
        elif self._is_weak_and_intervening_bid():
            bid = Pass('4014')
        elif self.opener_bid_one.name == '1NT':
            bid = self._overcaller_doubled_one_nt()
        else:
            bid = self._overcaller_has_doubled_weak()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _overcaller_doubled_one_nt(self) -> Call:
        """Bid after overcaller has doubled opener's 1NT."""
        if 4 <= self.hcp <= 7:
            bid = self._long_suit_bid()
        elif not self.responder_bid_one.is_pass:
            bid = self._overcaller_doubled_one_nt_responder_bid()
        elif self.five_card_suit_or_better and self.hcp <= 6:
            suit = self.longest_suit
            bid = self.next_level_bid(suit, '4015')
        elif self.hcp >= 7:
            bid = Pass('4129')
        else:
            bid = Pass('4016')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _overcaller_doubled_one_nt_responder_bid(self) -> Call:
        """Bid after overcaller has doubled opener's 1NT, responder has bid."""
        if self.is_semi_balanced and self.stoppers_in_bid_suits:
            bid = self._semi_balanced_stoppers()
        else:
            bid = self._long_suit_bid()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _semi_balanced_stoppers(self) -> Call:
        """Return bid if semi-balanced with stoppers in bid suits."""
        if self.hcp <= 9 and self.nt_level <= 2:
            bid = self.nt_bid(2, '4017')
        elif self.nt_level <= 3:
            bid = self.nt_bid(3, '4018')
        else:
            bid = self._long_suit_bid()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _overcaller_has_doubled_weak(self) -> Call:
        """Bid after partner has doubled with a weak hand."""
        if self._weak_and_responder_has_bid():
            bid = Pass('4019')
        elif self._leave_double():
            bid = Pass('4020')
        elif self.is_balanced and self.stoppers_in_bid_suits:
            bid = self._overcaller_has_doubled_weak_balanced()
        else:
            bid = self._long_suit_bid()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _overcaller_has_doubled_weak_balanced(self) -> Call:
        """Bid after partner has doubled with a weak balanced hand."""
        major_suit = self._find_four_card_major()
        if self.opponents_at_game:
            bid = Pass('4021')
        elif self.hcp <= 5:
            bid = self._long_suit_bid()
        elif major_suit:
            raise_level = self._find_level_for_major_raise(major_suit)
            bid = self.next_level_bid(major_suit, '4022', raise_level)
        elif self.opener_bid_one.level == 2 and self.nt_level <= 2:
            bid = self.nt_bid(2, '4023')
        elif self._has_ten_points_and_opening_bid_is_weak_three():
            bid = self.nt_bid(3, '4024')
        elif (self.hcp >= 12 and self.is_balanced and
              self.stoppers_in_bid_suits and self.nt_level <= 3):
            bid = self.nt_bid(3, '4025')
        elif self.hcp >= 10 and self.nt_level <= 2:
            bid = self.nt_bid(2, '4026')
        elif self.hcp >= 6 and self.nt_level <= 1:
            bid = self.nt_bid(1, '4027')
        else:
            bid = self._long_suit_bid()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _support_for_overcaller(self) -> Call:
        """Returns bid in partner's suit relevant to points."""
        if self._has_thirteen_points_and_five_card_major():
            bid = self._long_suit_bid()
        elif self.overcaller_bid_one.is_nt:
            bid = self._bid_after_nt()
        elif self.overcaller_bid_one.level == 1:
            bid = self._support_for_overcaller_level_one()
        elif self.overcaller_bid_one.level == 2:
            bid = self._support_for_overcaller_level_two()
        elif self.hcp >= 12 and not self.bidding_above_game:
            bid = self.next_level_bid(self.overcaller_suit_one, '4028')
        elif self._has_strength_and_support_to_overcall_three_level_bid():
            bid = self.next_level_bid(self.overcaller_suit_one, '4029')
        elif self.overcaller_bid_one.is_game:
            bid = Pass('4030', True)
        elif self.hcp <= 12 and self.next_level(self.overcaller_suit_one) >= 4:
            bid = Pass('4149', True)
        else:
            bid = Pass('4138', True)
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _bid_after_nt(self) -> Call:
        """ Return bid after overcaller bids NT."""
        if self.hcp <= 8:
            bid = Pass('4121')
        else:
            bid = Pass('4122')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _support_for_overcaller_level_one(self) -> Call:
        """Returns bid in partner's suit after level one bid."""
        suit_to_bid = self.overcaller_bid_one.denomination
        if self._has_ten_points_overcaller_bid_minor():
            bid = self._bid_nt()
        elif self._can_support_overcaller_at_given_level(level=2):
            bid = self.suit_bid(2, suit_to_bid, '4031', True)
        elif self._can_support_overcaller_at_given_level(level=3):
            bid = self.suit_bid(3, suit_to_bid, '4032', True)
        elif self._can_support_overcaller_at_given_level(level=4):
            bid = self.suit_bid(4, suit_to_bid, '4033', True)
        elif self.overcallers_suit_holding <= 3:
            bid = Pass('4123', True)
        elif self.hcp <= 6 or (self.hcp <= 12 and self.nt_level >= 3):
            bid = Pass('4148', True)
        else:
            bid = Pass('4034', True)
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _bid_nt(self) -> Call:
        """Return a NT bid."""
        if self.hcp >= 13 and self.nt_level <= 3:
            bid = self.nt_bid(3, '4035')
        elif self.hcp >= 12 and self.nt_level <= 2:
            bid = self.nt_bid(2, '4036')
        elif (self.is_semi_balanced
                and self.stoppers_in_bid_suits
                and self.nt_level <= 1):
            bid = self.nt_bid(1, '4037')
        elif self.suit_holding[self.overcaller_suit_one] <= 3:
            bid = Pass('4153')
        else:
            bid = Pass('4038')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _support_for_overcaller_level_two(self) -> Call:
        """Returns bid in partner's suit after level two bid."""
        suit_to_bid = self.overcaller_bid_one.denomination
        hand_value_points = self._get_hand_value_points(suit_to_bid)
        bid_level = self.next_level(suit_to_bid)
        if self._opening_hand_and_overcaller_bid_minor():
            bid = self.nt_bid(3, '4039')
        elif self._ten_points_and_overcaller_bid_minor():
            bid = self.nt_bid(2, '4040')
        elif hand_value_points in range(9, 14) and bid_level <= 3:
            if self.overcaller_has_jumped:
                bid_level = 4
            else:
                bid_level = 3
            bid = self.suit_bid(bid_level, suit_to_bid, '4041', True)
        elif self._good_support_for_overcaller():
            bid = self.suit_bid(4, suit_to_bid, '4042', True)
        elif self.hcp >= 13 and bid_level <= 4:
            bid = self.suit_bid(4, suit_to_bid, '4043', True)
        # elif hand_value_points >= 11 and bid_level <= 4:
        elif self.overcallers_suit_holding >= 5 and bid_level <= 4:
            bid = self.suit_bid(bid_level, suit_to_bid, '4044', True)
        elif self.overcallers_suit_holding >= 4 and bid_level <= 3:
            bid = self.suit_bid(bid_level, suit_to_bid, '4126', True)
        elif self._is_weak_with_some_support_for_overcaller():
            bid = self.suit_bid(bid_level, suit_to_bid, '4045', True)
        elif self._is_weak_with_some_support_for_overcaller():
            bid = self.suit_bid(bid_level, suit_to_bid, '4045', True)
        elif self.nt_level >= 3 and self.hcp <= 10:
            bid = Pass('4127', True)
        elif (self.hcp <= 12 and self.suit_holding[suit_to_bid] <= 3):
            bid = Pass('4147', True)
        elif self.opponents_at_game:
            bid = Pass('4154', True)
        elif self.next_level(suit_to_bid) >= 4:
            bid = Pass('4155', True)
        else:
            bid = Pass('4046', True)
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_stoppers(self) -> Call:
        """Return bid in NT relevant to points."""
        if self.overcaller_bid_one.level == 1:
            bid = self._no_support_stoppers_level_one()
        elif self.overcaller_bid_one.level == 2:
            bid = self._no_support_stoppers_level_two()
        else:
            bid = Pass('4047')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_stoppers_level_one(self) -> Call:
        """Return bid in NT relevant to points at level 1."""
        level = self.nt_level
        # Stoppers already checked
        if self.hcp in range(8, 12) and level <= 1:
            bid = self.nt_bid(1, '4048')
        elif self.hcp in range(12, 16) and level <= 2:
            bid = self.nt_bid(2, '4049')
        elif self.hcp >= 16 and level <= 3:
            bid = self.nt_bid(3, '4050')
        elif self.hcp <= 8 and self.nt_level >= 3:
            bid = Pass('4051')
        else:
            bid = Pass('4134')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _no_support_stoppers_level_two(self) -> Call:
        """Return bid in NT relevant to points at level 2."""
        level = self.nt_level
        if self.hcp in range(9, 14) and level <= 2:
            bid = self.nt_bid(2, '4052')
        elif self.hcp >= 14 and level <= 3:
            bid = self.nt_bid(3, '4053')
        elif self.hcp <= 15 and self.shape[0] <= 6 and self.nt_level >= 2:
            bid = Pass('4054')
        elif (self.hcp < 16
                and self.suit_holding[self.overcaller_suit_one] <= 4):
            bid = Pass('4161')
        else:
            bid = Pass('4135')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _long_suit_bid(self) -> Call:
        """Return bid based on own long suit."""
        if self.overcaller_bid_one.is_double:
            bid = self._bid_after_double()
        elif self.suit_length(self.longest_suit) >= 6 and self.hcp >= 11:
            bid = self._long_suit_after_suit()
        elif self._has_eleven_points_and_can_bid_nt():
            nt_level = self.nt_level
            if self._can_bid_five_card_major_below_nt_level():
                bid = self._long_suit_after_suit()
            else:
                if self.hcp >= 17 and self.nt_level == 1:
                    nt_level += 2
                if self.hcp >= 15 and self.nt_level == 2:
                    nt_level += 2
                elif self.hcp >= 13 and self.nt_level == 1:
                    nt_level += 1
                nt_level = min(3, nt_level)
                bid = self.nt_bid(nt_level, '4055')
        elif self._has_five_card_suit_and_seven_points():
            bid = self._long_suit_after_suit()
        elif self._has_seven_card_suit_and_shortage_in_overcallers_suit():
            bid = self.next_level_bid(self.longest_suit, '4056')
        elif self.hcp >= 6 and self.overcallers_suit_holding >= 3:
            bid = self.next_level_bid(
                self.overcaller_bid_one.denomination, '4057')
        elif self.suit_holding[self.overcaller_bid_one.denomination] <= 2:
            bid = Pass('4058')
        else:
            bid = Pass('4142')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _bid_after_double(self) -> Call:
        """Show long suit after overcaller has doubled."""
        suit = self._select_suit_after_double()
        (level, comment) = self._bid_after_double_get_level(suit)
        if self.partner_doubled_game:
            bid = Pass('4059')
        elif self._strong_in_openers_suit():
            bid = self._hold_openers_suit()
        elif self._responder_has_passed_after_one_nt_opening():
            bid = Pass('4060')
        elif self._suit_is_minor_and_can_bid_three_nt(suit, level):
            bid = self.next_nt_bid('4061')
        elif self._suit_is_minor_and_cannot_bid_three_nt(level):
            bid = Double('4062')
        elif self._has_opening_hand_and_can_bid_to_game(suit):
            bid = self.bid_to_game(suit, '4064')
        elif self._has_opening_hand_and_can_bid_to_game(self.longest_suit):
            bid = self.bid_to_game(self.longest_suit, '4064')
        elif self._overcaller_has_doubled_after_weak_two():
            if ((self.hcp >= 10 or
                    self.shape[0] >= 6) and
                    self.next_level(suit) <= 3):
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(suit, '4065', raise_level=raise_level)
        elif self._suit_bid_allowed(level, suit):
            bid = self.suit_bid(level, suit, comment)
        elif self._distributional_strong_hand(suit):
            if self.suit_holding[suit] >= 4:
                bid = self.next_level_bid(suit, '4141')
            elif self.stoppers_in_bid_suits:
                bid = self.next_level_nt('4140')
            else:
                bid = Pass('4066')
        elif self.last_bid.is_pass and suit not in self.opponents_suits:
            bid = self.next_level_bid(suit, '4067')
        elif (self._overcaller_has_doubled_one_nt() and
              self.right_hand_bid.is_value_call and
              self.shape[0] >= 5):
            bid = self.next_level_bid(self.longest_suit, '4068')
        elif self.long_suit in self.opponents_suits:
            bid = Pass('4069')
        else:
            bid = Pass('4069')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _hold_openers_suit(self) -> Call:
        """Return bid when holding opener's suit."""
        suit = self._select_suit_after_double()
        if self._is_balanced_and_holds_openers_suit():
            bid = self.next_nt_bid('4070')
        elif self._has_unbid_five_four():
            if self.opener_bid_one.level == 2:
                raise_level = 1
            else:
                raise_level = 0
            bid = self.next_level_bid(suit, '4071', raise_level)
        elif self.bid_history[-1] != 'P' and self.hcp <= 10:
            bid = Pass('4072')
        elif self.hcp <= 12:
            bid = Pass('4151')
        elif (self.overcaller_bid_one.is_double and self.hcp >= 10 and
              len(self.bid_history[-1]) >= 2 and
              self.cards_by_suit[self.bid_history[-1][1]]):
            bid = Pass('4156')
        elif self.overcaller_bid_one.is_double and self.hcp >= 15:
            bid = self.next_level_bid(suit, '4131', raise_level=1)
        elif (self.opener_bid_one.level in [2, 3] and
              self.overcaller_bid_one.is_double and
              self.hcp >= 10 and
              self.cards_by_suit[self.opener_suit_one.name][0].is_honour):
            bid = Pass('4158')
        else:
            # strength in opener's suit already established
            bid = Pass('4131')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _long_suit_after_suit(self) -> Call:
        """Show long suit after overcaller has bid suit."""
        if self._cannot_bid_at_this_level():
            bid = Pass('4073')
        elif (self.longest_suit.is_minor
                and self.is_balanced and self.nt_level <= 2):
            bid = self.nt_bid(2, '4074')
        elif self._is_strong_with_good_six_card_major():
            bid = self.bid_to_game(self.longest_suit, '4075')
        elif self._is_strong_with_good_seven_card_minor():
            bid = self.bid_to_game(self.longest_suit, '4076')
        elif self._has_support_for_overcaller_and_better_points():
            bid = self.next_level_bid(
                self.overcaller_bid_one.denomination, '4077')
        elif self._can_bid_own_suit_after_suit():
            if self.hcp >= 15:
                bid = self.next_level_bid(
                    self.longest_suit, '4078', raise_level=1)
            else:
                bid = self.next_level_bid(self.longest_suit, '4079')
        elif self._is_balanced_after_overcallers_minor():
            # Includes point count
            bid = self.next_nt_bid('4080')
        elif self.hcp >= 11 and self.overcallers_suit_holding >= 3:
            bid = self.next_level_bid(
                self.overcaller_bid_one.denomination, '4081')
        elif (self.overcaller_bid_one.level >= 2 and
              self.hcp >= 6 and
              self.shape[0] >= 6 and
              self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '4082')
        elif self.suit_holding[self.overcaller_bid_one.denomination] <= 2:
            bid = Pass('4083')
        elif self.previous_bid.is_pass:
            bid = Pass('4150')
        else:
            bid = Pass('4143')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_nt(self) -> Call:
        """Respond to a NT overcall."""
        if self.is_semi_balanced:
            bid = self._respond_to_nt_semi_balanced()
        elif self._has_ten_points_and_can_bid_six_card_major():
            bid = self.bid_to_game(self.longest_suit, '4084')
        elif self._has_six_points_and_can_bid_five_card_major():
            bid = self.next_level_bid(self.five_card_major_suit, '4085')
        else:
            bid = self._respond_to_nt_distributional()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_nt_semi_balanced(self) -> Call:
        """Respond to a NT overcall is semi_balanced."""
        if self.overcaller_bid_one.name == '1NT':
            bid = self._respond_to_one_nt_semi_balanced()
        elif self.overcaller_bid_one.name == '2NT':
            bid = self._respond_to_two_nt_semi_balanced()
        else:
            bid = self._respond_to_two_nt_semi_balanced_weak()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_one_nt_semi_balanced(self) -> Call:
        """Respond to 1NT overcall is semi_balanced."""
        if self._can_bid_stayman_after_overcallers_one_nt():
            bid = self.club_bid(2, '4086')
        elif self._can_bid_stayman_after_overcallers_two_nt():
            bid = self.club_bid(3, '4087')
        elif 8 <= self.hcp <= 9 and self.nt_level <= 2:
            bid = self.nt_bid(2, '4088')
        elif self._has_six_card_major_and_can_bid_game() and self.hcp >= 9:
            bid = self.suit_bid(4, self.longest_suit, '4089')
        elif self._has_nine_points_and_four_card_major():
            bid = self.next_level_bid(self.longest_suit, '4090', raise_level=1)
        elif self.hcp >= 9 and self.nt_level <= 3:
            bid = self.nt_bid(3, '4091')
        elif self.hcp <= 8:
            bid = Pass('4128')
        else:
            bid = Pass('4092')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_two_nt_semi_balanced(self) -> Call:
        """Respond to 2NT overcall is semi_balanced."""
        if self._has_eight_points_and_five_card_major():
            bid = self.next_level_bid(self.five_card_major_suit, '4093')
        elif self._overcaller_bid_nt_and_four_card_major():
            bid = self.suit_bid(3, self.club_suit, '4094')
        elif self.hcp >= 8 and self.nt_level <= 3:
            bid = self.nt_bid(3, '4095')
        else:
            bid = self._respond_to_two_nt_semi_balanced_weak()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_two_nt_semi_balanced_weak(self) -> Call:
        """Respond to 2NT overcall is semi_balanced with weak hand."""
        if (self.hcp >= 4
                and self.nt_level <= 3
                and self.opener_bid_one.level == 1):
            bid = self.nt_bid(3, '4096')
        elif self.hcp <= 10 and self.opener_bid_one.level >= 2:
            bid = Pass('4097')
        elif self.hcp <= 3:
            bid = Pass('4146')
        else:
            bid = Pass('4145')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_nt_distributional(self) -> Call:
        """Respond to a NT overcall if distributional."""
        if (self.overcaller_bid_one.name == '2NT' and
                self.opener_bid_one.level == 1):
            bid = self._respond_to_nt_distributional_overcaller_jumped()
        elif self.hcp >= 4:
            bid = self._respond_to_nt_after_weak_2()
        else:
            bid = Pass('4098')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_nt_after_weak_2(self) -> Call:
        """Return bid after opener bid weak 2"""
        suit = self._select_suit()
        bid = self.next_level_bid(suit)
        if (bid.is_minor and bid.level == 2
                and self.stoppers_in_bid_suits
                and self.hcp >= 8):
            bid = self._respond_to_nt_distributional_with_clubs()
        elif (self.suit_length(suit) >= 5 and
                suit not in self.opponents_suits and
                self.next_level(suit) < self.bid_to_game(suit).level and
                suit != self.club_suit):
            bid = self.next_level_bid(suit, '4099')
        elif self._is_five_four_or_five_five() and self.hcp >= 10:
            suit = self.longest_suit
            if suit in self.opponents_suits:
                suit = self.second_suit
            if self.next_level(suit) < self.bid_to_game(suit).level:
                bid = self.next_level_bid(suit, '4100')
            elif self.opponents_at_game:
                bid = Pass('4137')
            else:
                bid = Pass('4101')
        elif (self.opener_bid_one.level == 2
                and self.overcaller_bid_one.name == '2NT'):
            bid = self._nt_after_weak_opening()
        elif self.hcp <= 8:
            bid = Pass('4102')
        elif self.overcaller_bid_one.is_nt and self.shape[3] <= 1:
            bid = Pass('4157')
        else:
            bid = Pass('4136')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _nt_after_weak_opening(self) -> Call:
        """Return bid after overcaller bids NT after weak 2 opening."""
        if self.hcp <= 6 and self.bid_history[-1] == 'P':
            bid = Pass('4103')
        elif (self.hcp >= 8
                and self.four_card_major_suit
                and self.next_level(self.club_suit) <= 3):
            bid = self.next_level_bid(self.club_suit, '4104')
        elif self.shape[0] >= 6 and self.hcp >= 5 and self.nt_level <= 3:
            bid = self.next_level_bid(self.longest_suit, '4132')
        elif self.hcp >= 9 and self.nt_level <= 3:
            bid = self.nt_bid(3, '4105')
        elif self.opponents_at_game:
            bid = Pass('4106')
        else:
            bid = Pass('4139')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_nt_distributional_with_clubs(self) -> Call:
        """Respond to a NT overcall if best suit is clubs."""
        if self.hcp >= 10:
            bid = self.nt_bid(3, '4133')
        elif self.hcp >= 8:
            bid = self.nt_bid(2, '4107')
        else:
            bid = Pass('4108')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _respond_to_nt_distributional_overcaller_jumped(self) -> Call:
        """Respond to a jump NT overcall is distributional."""
        if self._can_bid_five_card_major_at_level_three():
            bid = self.suit_bid(3, self.longest_suit, '4109')
        elif self.hcp <= 9:
            bid = Pass('4110')
        else:
            bid = Pass('4130')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    # Various utility functions

    def _fourth_suit_bid(self, suit: Suit) -> bool:
        """Return True if suit is fourth suit bid."""
        value = False
        suits = [0, 0, 0, 0]
        suits[suit.rank] = 1
        for bid in self.bid_history:
            if Bid(bid).is_suit_call:
                bid_suit = Bid(bid).denomination.rank
                suits[bid_suit] = 1
        if sum(suits) == 4:
            value = True
        return value

    def _select_suit(self, force_major=False) -> Suit:
        """Return best suit to bid."""
        (suit_one, suit_two) = self._top_two_unbid_suits()
        suit_one_holding = self.suit_length(suit_one)
        suit_two_holding = self.suit_length(suit_two)
        if (suit_one_holding == suit_two_holding or
                (suit_one_holding == 5 and suit_two_holding == 4 and
                 suit_one.is_minor and suit_two.is_major)):
            if suit_two.is_major:
                suit = self.cheaper_suit(suit_one, suit_two)
            elif suit_one.is_major:
                suit = suit_one
            elif (suit_one.is_major and
                  suit_one_holding == suit_two_holding):
                suit = suit_one
                if not force_major and suit_one_holding == suit_two_holding:
                    suit = self.cheaper_suit(suit_one, suit_two)
            else:
                suit = self.cheaper_suit(suit_one, suit_two)
        else:
            suit = suit_one
        if (suit == self.opener_suit_one
                or suit == self.responder_bid_one.denomination):
            if suit_two_holding >= 0:
                suit = suit_two
            else:
                suit = self.no_trumps
        return suit

    def _top_two_unbid_suits(self) -> tuple[Suit]:
        """Return two longest unbid suits."""
        suit_index = 0
        suit_offset = 1
        found = False
        # Loop through suits to find two longest unbid suits.
        suit_one, suit_two = None, None
        while not found:
            suit_one = self.ordered_holding[suit_index][1]
            suit_two = self.ordered_holding[suit_index+suit_offset][1]
            if (suit_one == self.opener_suit_one or
                    suit_one == self.responder_bid_one.denomination):
                suit_index += 1
            elif (suit_two == self.opener_suit_one or
                  suit_two == self.responder_bid_one.denomination):
                suit_offset += 1
            else:
                found = True
        return (suit_one, suit_two)

    def _can_bid_to_game(self, suit: Suit) -> bool:
        """Return True if conditions are met for raise to game."""
        sufficient_points = (self.hcp >= 8 or
                             (self.hcp >= 7 and self.shape[3] <= 1))
        if (self.suit_length(suit) >= 2 and
                sufficient_points and
                not self.is_insufficient_bid(self.bid_to_game(suit)) and
                (self.opener_bid_one.level == 1 or
                 suit.is_major)):
            return True
        return False

    def _find_level_for_major_raise(self, major_suit: Suit) -> int:
        """Return the appropriate raise level for major suit  support."""
        if self.next_level(major_suit) >= 4:
            raise_level = 0
        elif self.hcp >= 13:
            raise_level = 2
        elif self.hcp >= 10:
            raise_level = 1
        else:
            raise_level = 0
        return raise_level

    def _find_four_card_major(self) -> Suit:
        """Return an unbid four card major."""
        if self.spades >= 4 and self.spade_suit not in self.opponents_suits:
            suit = self.spade_suit
        elif (self.hearts >= 4 and
              self.heart_suit not in self.opponents_suits):
            suit = self.heart_suit
        else:
            suit = None
        return suit

    def _leave_double(self) -> bool:
        suit = self._select_suit()
        openers_suit = self.opener_bid_one.denomination
        if (self.opener_bid_one.level > 1 and
            self.opener_bid_one.is_suit_call and
            suit.is_minor and
                self.suit_length(openers_suit) >= 4 and
                self.hcp >= 10 and
                self.responder_bid_one.is_pass):
            return True
        return False

    def _get_hand_value_points(self, suit: Suit) -> int:
        """Return hand value points."""
        suit_cards = self.suit_length(suit)
        hand_value_points = (self.hcp +
                             self.support_shape_points(suit) +
                             suit_cards - 3)
        if self.shape[3] == 0:
            hand_value_points += 1
        return hand_value_points

    def _select_suit_after_double(self) -> Suit:
        five_four = (self.five_four and
                     (self.longest_suit in self.opponents_suits or
                      self.second_suit in self.opponents_suits))
        if self.five_five:
            if (self.longest_suit.rank < self.second_suit.rank and
                    self.second_suit not in self.opponents_suits):
                return self.second_suit
            if self.longest_suit not in self.opponents_suits:
                return self.longest_suit
            return self.second_suit
        elif five_four:
            if self.longest_suit not in self.opponents_suits:
                return self.longest_suit
            return self.second_suit
        return self._select_suit(force_major=True)

    def _suit_bid_allowed(self, level: int, suit: Suit) -> bool:
        """Return True if a suit bid is allowed."""
        double_forcing = Bid(self.bid_history[-1]).is_pass
        sufficient_points = (self.hcp >= 9 or level <= 2)
        level_low_enough = (self.next_level(suit) <= 4
                            and level >= self.next_level(suit))

        if not double_forcing and self.suit_length(suit) < 4:
            return False
        if double_forcing and suit not in self.opponents_suits:
            return True
        if (sufficient_points and
                level_low_enough and
                suit not in self.opponents_suits):
            return True
        return False

    def _bid_after_double_get_level(self, suit: Suit) -> tuple[int, str]:
        """Return level and comment with a long suit after a double."""
        if self.opener_bid_one.level == 1:
            (level, comment) = self._bid_after_double_level_one(suit)
            if level:
                return (level, comment)
            return (self.next_level(suit), '4111')
        if self.opener_bid_one.level == 2:
            return self._bid_after_double_level_two(suit)
        if self.opener_bid_one.level == 3:
            return self._bid_after_double_level_three(suit)

        if self.opener_bid_one.level == 4:
            return (self.next_level(suit), '4117')
        return (self.next_level(suit), '4118')

    def _bid_after_double_level_one(self, suit: Suit) -> tuple[int, str]:
        """Return level and comment with a
        long suit after a double at level 1."""
        level = self.next_level(suit)
        if self.responder_bid_one.is_value_call:
            if self.hcp >= 12:
                level += 1
            return (level, '4112')
        if (self.hcp >= 13 and suit.is_major and self.suit_length(suit) >= 5):
            return (4, '4114')
        if self.hcp >= 12 and not self.stoppers_in_bid_suits:
            return (level + 2, '4119')
        if self.hcp >= 9 and level == 1:
            return (level + 1, '4120')
        return (level, '4113')

    def _bid_after_double_level_two(self, suit: Suit) -> tuple[int, str]:
        """Return level and comment with a
        long suit after a double at level 2."""
        level = self.next_level(suit)
        level = max(2, level)
        if self.hcp >= 8:
            level = max(3, level)
        return (level, '4115')

    def _bid_after_double_level_three(self, suit: Suit) -> tuple[int, str]:
        """Return level and comment with a
        long suit after a double at level 3."""
        level = self.next_level(suit)
        level = max(3, level)
        if self.hcp >= 8:
            level = max(4, level)
        return (level, '4116')

    # Various boolean functions

    def _overcall_bid_at_three_level(self) -> bool:
        """Return True if overcaller has bid at level 3."""
        return (self.overcaller_bid_one.level >= 3 and
                self.opener_bid_one.level != 3)

    def _has_biddable_eight_card_suit(self) -> bool:
        """Return True if with biddable 8 card suit."""
        next_level = self.next_level(self.longest_suit)
        return (self.shape[0] == 8 and
                next_level <= self.longest_suit.game_level and
                self.longest_suit not in self.opponents_suits)

    def _has_three_card_support_for_overcallers_major(self) -> bool:
        """Return True if 3 card support for overcaller's major."""
        return (self.overcaller_bid_one.is_major and
                self.overcallers_suit_holding >= 3)

    def _is_weak_no_support_for_overcaller(self) -> bool:
        """Return True if weak and no support for overcaller."""
        return (self.bid_history[-1] == 'P' and
                8 <= self.hcp <= 9 and
                self.overcallers_suit_holding <= 2 and
                self.shape[0] <= 5 and
                self.nt_level == 1 and
                self.stoppers_in_bid_suits)

    def _has_biddable_suit(self) -> bool:
        """Return True if hand has a biddable suit."""
        biddable_six_card_suit = (self.shape[0] >= 6 and
                                  self.overcallers_suit_holding <= 2)
        can_compete_after_nt = (self.opener_bid_one.name == '1NT' and
                                self.overcaller_bid_one.is_suit_call and
                                self.overcaller_bid_one.level == 2 and
                                self.hcp >= 10)
        longest_suit = self.suit_length(self.longest_suit)
        if (longest_suit == self.suit_length(self.second_suit) and
                self.second_suit == self.overcaller_bid_one.denomination):
            my_suit = self.second_suit
        else:
            my_suit = self.longest_suit
        has_own_five_card_suit = (self.overcallers_suit_holding <= 1 and
                                  self.shape[0] >= 5)
        has_own_five_card_major = (self.hcp >= 8
                                   and self.shape[0] >= 5
                                   and my_suit.is_major)
        can_bid_own_suit = (self.hcp >= 8 and
                            self.shape[0] >= 5 and
                            self.responder_bid_one.is_value_call and
                            self.overcaller_bid_one.is_minor and
                            my_suit.is_major)

        return (biddable_six_card_suit or
                can_bid_own_suit or
                can_compete_after_nt or
                has_own_five_card_major or
                has_own_five_card_suit)

    def _can_bid_nt_opposition_has_no_fit(self) -> bool:
        """Return True if opposition weak or no fit can bid nt."""
        return (self.stoppers_in_bid_suits and
                (not self.responder_bid_one.is_pass or
                 self.hcp >= 16) and
                (self.opener_suit_one != self.responder_bid_one.denomination or
                 self.responder_bid_one.level <= 2))

    def _can_support_overcaller(self) -> bool:
        """Return True if can support overcaller."""
        if self.overcaller_has_jumped:
            support_for_overcaller = 2
        else:
            support_for_overcaller = 3
        return self.overcallers_suit_holding >= support_for_overcaller

    def _has_opening_values_can_bid_nt(self) -> bool:
        """Return True if opening values and can bid nt."""
        return (self.is_balanced and
                self.hcp >= 11 and
                not self.responder_bid_one.is_value_call and
                self.stoppers_in_bid_suits)

    def _can_bid_six_card_suit_at_three_level(self) -> bool:
        """Return True if can bid 6 card suit at level 3."""
        return (self.shape[0] >= 6 and
                self.hcp >= 10 and
                self.next_level(self.longest_suit) <= 3 and
                self.longest_suit not in self.opponents_suits)

    def _has_ten_points_and_stoppers(self) -> bool:
        """Return True if 10 points and can bid 3NT."""
        return (self.hcp >= 10 and
                self.overcaller_bid_one.denomination.is_minor and
                self.stoppers_in_bid_suits and
                self.nt_level == 3)

    def _has_eight_points_and_stoppers(self) -> bool:
        """Return True if eight points and stoppers."""
        return (self.hcp >= 8 and
                self.stoppers_in_unbid_suits() and
                self.nt_level == 3)

    def _has_eleven_points_balanced_and_responder_has_bid_game(self) -> bool:
        """Return True if 11 points and responder at game."""
        return (self.hcp >= 11 and
                self.shape[0] == 4 and
                self.responder_bid_one.is_game)

    def _is_weak_and_intervening_bid(self) -> bool:
        """Return True if weak after overcaller has doubled."""
        double_index = self.bid_history.index('D')
        next_bid = Bid(self.bid_history[double_index+1])
        return (next_bid.is_suit_call and
                self.stoppers_in_bid_suits and
                self.hcp <= 5 and
                self.shape[0] <= 5)

    def _weak_and_responder_has_bid(self) -> bool:
        """Return True if weak and responder has bid."""
        return (self.hcp <= 5 and
                not self.responder_bid_one.is_pass and
                self.responder_bid_one.level <= 2)

    def _has_ten_points_and_opening_bid_is_weak_three(self) -> bool:
        """Return True if 10+ points and opener has bid at 3 level."""
        return (self.opener_bid_one.level == 3 and
                self.nt_level <= 3 and
                self.hcp >= 10)

    def _has_thirteen_points_and_five_card_major(self) -> bool:
        """Return True if 13 points and five card major."""
        return (self.overcaller_bid_one.denomination.is_minor and
                self.five_card_major_or_better and
                self.hcp >= 13)

    def _has_ten_points_overcaller_bid_minor(self) -> bool:
        """Return True if 10 points and overcaller bid minor."""
        return (self.overcaller_bid_one.is_minor and
                self.is_semi_balanced and
                self.stoppers_in_bid_suits and
                self.hcp >= 10 and
                not Bid(self.bid_history[-1]).is_value_call)

    def _can_support_overcaller_at_given_level(self, level: int) -> bool:
        """Return True if can support overcaller at level 2."""
        denomination = self.overcaller_bid_one.denomination
        hand_value_points = self._get_hand_value_points(denomination)
        if level == 2:
            points_range = range(8, 13)
        elif level == 3:
            points_range = range(12, 16)
        elif level == 4:
            points_range = range(16, 40)
        else:
            points_range = range(0, 0)
        return (hand_value_points in points_range and
                self.next_level(denomination) <= level)

    def _opening_hand_and_overcaller_bid_minor(self) -> bool:
        """Return True if opening hand and overcaller bids minor."""
        return (self.hcp >= 13 and
                self.is_semi_balanced and
                self.overcaller_bid_one.denomination.is_minor and
                self.stoppers_in_bid_suits and
                self.nt_level <= 3)

    def _ten_points_and_overcaller_bid_minor(self) -> bool:
        """Return True if 10+ points and overcaller bids minor."""
        return (self.hcp >= 10 and
                self.overcaller_bid_one.denomination.is_minor and
                self.stoppers_in_bid_suits and
                self.is_semi_balanced and
                self.nt_level <= 2)

    def _good_support_for_overcaller(self) -> bool:
        """Return True if good support for overcaller."""
        denomination = self.overcaller_bid_one.denomination
        hand_value_points = self._get_hand_value_points(denomination)
        return (self.overcaller_bid_one.level <= 4 and
                self.next_level(denomination) <= 4 and
                (self.hcp >= 16 or
                 (hand_value_points >= 14 and
                  self.overcallers_suit_holding >= 5)))

    def _is_weak_with_some_support_for_overcaller(self) -> bool:
        """Return True if weak with some support for overcaller."""
        return (self.overcallers_suit_holding >= 4 and
                self.suit_points(self.overcaller_bid_one.denomination) >= 1 and
                self.hcp >= 5 and
                self.opener_bid_one.level == 2)

    def _has_eleven_points_and_can_bid_nt(self) -> bool:
        """Return True if 11 points and can bid nt."""
        return (self.stoppers_in_bid_suits and
                self.hcp >= 11 and
                self.nt_level <= 3)

    def _can_bid_five_card_major_below_nt_level(self):
        """Return True if can bid 5 card major."""
        return (self.next_level(self.longest_suit) <= self.nt_level and
                self.shape[0] >= 5 and
                self.longest_suit.is_major)

    def _has_five_card_suit_and_seven_points(self) -> bool:
        """Return True if 7 points and 5 card suit."""
        suit = self._select_suit()
        return (suit.is_suit and
                self.suit_length(suit) >= 5 and
                self.hcp >= 7 and
                not self.opener_bid_two.is_nt)

    def _has_seven_card_suit_and_shortage_in_overcallers_suit(self) -> bool:
        """Return True if 7 card suit and shortage in overcallers suit."""
        return (self.overcallers_suit_holding <= 1 and
                self.shape[0] >= 7 and
                self.hcp >= 3 and
                self.next_level(self.longest_suit) <= 2 and
                self.longest_suit not in self.opponents_suits)

    def _strong_in_openers_suit(self) -> bool:
        """Return True if strong in openers suit."""
        return (self.opener_bid_one.is_suit_call and
                ((self.hcp >= 8 and
                  self.suit_length(self.opener_suit_one) >= 4) or
                 (self.hcp >= 6 and
                  self.suit_length(self.opener_suit_one) >= 5)))

    def _responder_has_passed_after_one_nt_opening(self) -> bool:
        """Return True if 7 points and responder passes after 1NT."""
        return (self.opener_bid_one.is_nt and
                self.hcp >= 7 and
                self.responder_bid_one.is_pass)

    def _suit_is_minor_and_can_bid_three_nt(self, suit, level: int) -> bool:
        """Return True if suit is minor and level over 3 NT."""
        return (10 <= self.hcp <= 12 and
                level >= 4 and
                suit.is_minor and
                self.stoppers_in_bid_suits)

    def _suit_is_minor_and_cannot_bid_three_nt(self, level: int) -> bool:
        """Return True if suit is minor and level over 3 NT."""
        return (self.hcp >= 8 and
                level >= 4 and
                self.responder_bid_one.denomination.is_minor and
                not self.overcaller_bid_one.is_double)

    def _has_opening_hand_and_can_bid_to_game(self, suit: Suit) -> bool:
        """Return True if opening hand and can bid to game in suit."""
        double_level = self.double_level()
        game_level = self.game_level(suit)
        next_level = self.next_level(suit)
        return (self.hcp >= 13 and
                double_level == 3 and
                next_level <= game_level)

    def _overcaller_has_doubled_after_weak_two(self) -> bool:
        """Return True if overcaller has doubled after weak two opening."""
        return (self.hcp >= 6 and
                self.opener_bid_one.level == 2 and
                self.longest_suit not in self.opponents_suits)

    def _distributional_strong_hand(self, suit: Suit) -> bool:
        """Return True if distributional hand."""
        return (self.hcp + self.distribution_points >= 12 and
                self.next_level(suit) <= 4 and
                suit not in self.opponents_suits)

    def _overcaller_has_doubled_one_nt(self):
        """Return True if overcaller has doubled 1NT."""
        return (self.opener_bid_one.name == '1NT' and
                self.overcaller_bid_one.is_double and
                self.hcp >= 6 and
                self.longest_suit not in self.opponents_suits)

    def _is_balanced_and_holds_openers_suit(self) -> bool:
        """Return True if balanced and holds openers suit."""
        return (self.hcp >= 8 and
                self.is_balanced and
                self.stoppers_in_bid_suits)

    def _has_unbid_five_four(self) -> bool:
        """Return True if unbid 5/4 suits."""
        return (self.hcp >= 8 and
                self.five_four and
                (self.longest_suit not in self.opponents_suits or
                 self.second_suit not in self.opponents_suits))

    def _cannot_bid_at_this_level(self) -> bool:
        """Return True if cannot make a sensible bid at this level."""
        level = self.next_level(self.longest_suit)
        jump = self.is_jump(self.opener_bid_one, self.overcaller_bid_one)
        cannot_bid_at_level_three = (level >= 3
                                     and self.hcp <= 11
                                     and self.shape[0] <= 4
                                     and not jump)
        cannot_bid_at_level_two = (level >= 2
                                   and self.hcp <= 9
                                   and self.shape[0] <= 5
                                   and not jump)
        return (level >= 7 or
                cannot_bid_at_level_three or
                cannot_bid_at_level_two or
                self.hcp < 6)

    def _is_strong_with_good_six_card_major(self) -> bool:
        """Return True if strong with good six card major."""
        return (self.shape[0] >= 6 and
                self.hcp >= 15 and
                self.suit_points(self.longest_suit) >= 8 and
                self.longest_suit.is_major and
                self.longest_suit not in self.opponents_suits)

    def _is_strong_with_good_seven_card_minor(self) -> bool:
        """Return True if strong with good 7 card minor."""
        return (self.shape[0] >= 7 and
                self.hcp >= 15 and
                self.suit_points(self.longest_suit) >= 8 and
                self.longest_suit.is_minor and
                self.longest_suit not in self.opponents_suits)

    def _can_bid_own_suit_after_suit(self) -> bool:
        """Return True if can bid own suit."""
        return ((self.next_level(self.longest_suit) <= 2 or
                 self.hcp >= 10 or
                 self.longest_suit.is_major) and
                (self.hcp >= 10
                 or not self._fourth_suit_bid(self.longest_suit))
                and self.longest_suit not in self.opponents_suits)

    def _has_six_points_and_can_bid_five_card_major(self) -> bool:
        """Return True if 6+ points and can bid 5 card major."""
        return (self.five_card_major_suit and
                self.five_card_major_suit not in self.opponents_suits and
                self.hcp >= 6 and
                self.next_level(self.longest_suit) <= 4)

    def _has_ten_points_and_can_bid_six_card_major(self) -> bool:
        """Return True if 10+ points and can bid 6 card major."""
        return (self.shape[0] >= 6 and
                self.longest_suit.is_major and
                self.longest_suit not in self.opponents_suits and
                self.hcp >= 10 and
                self.next_level(self.longest_suit) <= 4)

    def _can_bid_stayman_after_overcallers_one_nt(self) -> bool:
        """Return True if 4 card major over overcallers 1NT."""
        return (self.overcaller_bid_one.name == '1NT' and
                self.four_card_major_suit and
                self.suit_length(self.four_card_major_suit) == 4 and
                self.hcp >= 8 and
                self.next_level(self.club_suit) <= 2)

    def _can_bid_stayman_after_overcallers_two_nt(self) -> bool:
        """Return True if 4 card major over overcallers 2NT."""
        return (self.overcaller_bid_one.name == '2NT' and
                self.four_card_major_suit and
                self.hcp >= 6)

    def _has_six_card_major_and_can_bid_game(self) -> bool:
        """Return True if six_card_major."""
        return (self.shape[0] >= 6 and
                self.longest_suit.is_major and
                self.next_level(self.longest_suit) <= 4 and
                self.longest_suit not in self.opponents_suits)

    def _has_nine_points_and_four_card_major(self):
        """Return True if has 9 points and 4 card major."""
        return (self.five_card_major_or_better and
                self.hcp >= 9 and
                self.longest_suit not in self.opponents_suits)

    def _has_eight_points_and_five_card_major(self) -> bool:
        """Return True if 6 points and a five card major."""
        return (self.five_card_major_suit and
                self.five_card_major_suit not in self.opponents_suits and
                self.hcp >= 8)

    def _is_five_four_or_five_five(self) -> bool:
        """Return True if 5/4 or 5/5."""
        return ((self.five_four or self.five_five) and
                (self.longest_suit not in self.opponents_suits or
                 self.second_suit not in self.opponents_suits))

    def _can_bid_five_card_major_at_level_three(self) -> bool:
        """Return True if can bid 5 card major at level 3."""
        return (self.longest_suit.is_major and
                self.longest_suit not in self.opponents_suits and
                self.suit_length(self.longest_suit) >= 5 and
                self.next_level(self.longest_suit) == 3)

    def _is_balanced_after_overcallers_minor(self) -> bool:
        """Return True if balanced and stoppers in bid suit."""
        return (self.hcp >= 11 and
                self.is_balanced and
                self.overcaller_bid_one.denomination.is_minor and
                self.stoppers_in_bid_suits)

    def _has_strength_and_support_to_overcall_three_level_bid(self) -> bool:
        """Return True if 10 points and supporter for overcaller."""
        return (self.overcaller_bid_one.level == 3 and
                self.overcaller_bid_one.denomination.is_major and
                self.hcp >= 10 and
                self.overcallers_suit_holding >= 3 and
                self.next_level(self.overcaller_bid_one.denomination) <= 4)

    def _has_support_for_overcaller_and_better_points(self) -> bool:
        """Return True if support and better points in overcaller's suit."""
        return (self.overcallers_suit_holding >= 4 and
                self.suit_points(self.overcaller_bid_one.denomination) >= 5 and
                self.suit_points(self.longest_suit) <= 3)

    def _overcaller_bid_nt_and_four_card_major(self) -> bool:
        """Return True if overcall has bid no trumps and a 4 card major."""
        return (self.hcp >= 8 and
                self.longest_suit.is_major and
                self.next_level(self.club_suit) <= 3)



"""
    Bid for Game
    Acol opening bid module
"""
import inspect
from bridgeobjects import Board, Card
from bfgbidding.bidding import Bid, Pass, HandSuit
from bfgbidding.hand import Hand

inspection = inspect.currentframe


class OpeningBid(Hand):
    """BfG OpeningBid class."""
    def __init__(self, hand_cards: list[Card], board: Board):
        super(OpeningBid, self).__init__(hand_cards, board)
        self.trace = 0

    def suggested_bid(self) -> Bid:
        """Directs control to relevant method and return a Bid."""
        if self.hcp >= 23:
            bid = self.club_bid(2, '1001')
        elif self.is_balanced or (
                self.is_semi_balanced and 20 <= self.hcp <= 22):
            bid = self._balanced_openings()
        elif self._can_open_normally():
            bid = self._unbalanced_openings()
        elif self.shape[0] >= 6:
            bid = self._weak_opening_bid()
        else:
            bid = Pass('1002')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _can_open_normally(self) -> Bid:
        """Return True if hand suitable for opening bid."""
        longest_suit_points_points = self.suit_points(self.longest_suit)
        second_suit_points = self.suit_points(self.second_suit)
        if self.hcp >= 12:
            value = True
        elif (self.hcp == 11 and (self.shape[0] >= 6 or self.five_five) or
              (self.five_four and
               longest_suit_points_points + second_suit_points >= 11)):
            value = True
        else:
            value = False
        return value

    def _balanced_openings(self) -> Bid:
        """Return bid for balanced hands."""
        if 12 <= self.hcp <= 14:
            bid = self._weak_balanced_bid()
        elif 14 <= self.hcp < 20:
            bid = self._unbalanced_openings()
        elif 20 <= self.hcp <= 22:
            bid = self.nt_bid(2, '1003')
        else:
            bid = Pass('1004')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_balanced_bid(self) -> Bid:
        """Return bid for weak (12-14) balanced hands."""
        if self._five_card_major_with_seven_suit_points():
            bid = self._unbalanced_openings()
        else:
            bid = self.nt_bid(1, '1005')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _unbalanced_openings(self) -> Bid:
        """ return bid for unbalanced hands."""
        # points checked before entry to here
        if self.shape[0] >= 8 and self.hcp <= 14:
            bid = self._weak_eight_card_suits_bid()
        elif self.shape[0] >= 5:
            bid = self._bid_with_five_card_suit()
        elif self.shape == [4, 4, 4, 1]:
            bid = self._four_four_four_one_bid()
        else:
            bid = self._four_card_suits()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_eight_card_suits_bid(self) -> Bid:
        """Return bid for 8+ card suits and fewer than 15 points."""
        suit = self.longest_suit
        if suit.is_major:
            game_level = 4
        else:
            game_level = 5
        level = self.shape[0] - 4
        level = min(level, game_level)
        if self.longest_suit.is_minor and self.shape[1] >= 4:
            bid = self.suit_bid(1, self.longest_suit, '1006')
        else:
            bid = self.suit_bid(level, suit, '1007')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _four_card_suits(self) -> Bid:
        """Return bid with two four card suits."""
        if self.spades == 4:
            bid = self._four_spades()
        elif self.hearts == 4:
            bid = self.heart_bid(1, '1008')
        elif self.diamonds == 4:
            bid = self._four_diamonds()
        else:
            bid = self.club_bid(1, '1009')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _four_spades(self) -> Bid:
        """Return bid if hand has 4 spades."""
        if self.hearts == 4:
            bid = self.heart_bid(1, '1010')
        elif self.diamonds == 4:
            bid = self.spade_bid(1, '1011')
        elif self.clubs == 4:
            bid = self.club_bid(1, '1012')
        else:
            bid = self.spade_bid(1, '1013')  # 4,3,3,3
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _four_diamonds(self) -> Bid:
        """Return bid if hand has 4 diamonds (and possibly 4 clubs)."""
        if self.clubs == 4:
            bid = self.club_bid(1, '1014')
        else:
            bid = self.diamond_bid(1, '1015')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _bid_with_five_card_suit(self) -> Bid:
        """Return bid with a five card suit."""
        if not self.equal_long_suits:
            if self.is_balanced and self.hcp <= 14:
                bid = self.suit_bid(1, self.longest_suit, '1016')
            else:
                bid = self.suit_bid(1, self.longest_suit, '1017')
            if self.hcp <= 11 and self.shape[0] == 5 and self.shape[1] == 4:
                bid.call_id = '1018'
            elif self.hcp <= 11:
                bid.call_id = '1019'
        else:
            bid = self._five_five_hands_bid()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _five_five_hands_bid(self) -> Bid:
        """Return bid for five/five hands."""
        if self.spades >= 5 and self.clubs >= 5:
            bid = self.club_bid(1, '1020')
        elif self.spades >= 5:
            bid = self.spade_bid(1, '1021')
        elif self.hearts >= 5:
            bid = self.heart_bid(1, '1022')
        elif self.diamonds >= 5:
            bid = self.diamond_bid(1, '1023')
        else:
            assert False, 'Bid not defined'
        if self.hcp <= 11:
            bid.call_id = '1024'
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _four_four_four_one_bid(self) -> Bid:
        """Return bid for 4441 hands."""
        if self.clubs == 1:
            bid = self.heart_bid(1, '1025')
        elif self.diamonds == 1:
            bid = self.club_bid(1, '1026')
        elif self.hearts == 1:
            bid = self.diamond_bid(1, '1027')
        elif self.spades == 1:
            bid = self.heart_bid(1, '1028')
        else:
            assert False, 'Bid not defined'
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_opening_bid(self) -> Bid:
        """Return bid for weak opening hands."""
        if self.seven_six:
            return self._weak_seven_six_bid()
        elif self.six_six:
            return self._weak_six_six_bid()
        else:
            bid = self._select_weak_bid()
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _select_weak_bid(self) -> Bid:
        """Return weak bids."""
        if (self.hcp >= 6 and self.shape[0] == 6
                and self.longest_suit != self.club_suit):
            bid = self._weak_two_bid()
        elif self.shape[0] >= 7:
            bid = self._weak_three_bid()
        else:
            bid = Pass('1029')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_two_bid(self) -> Bid:
        """Return weak two bid."""
        suit = self.longest_suit
        suit_quality = HandSuit(suit, self).suit_quality()
        if suit != self.club_suit and suit_quality >= 0:
            bid = self.suit_bid(2, suit, '1030')
        else:
            bid = Pass('1031')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_three_bid(self) -> Bid:
        """Return weak three bid."""
        suit = self.longest_suit
        suit_quality = HandSuit(suit, self).suit_quality()
        if self.hcp <= 9 and suit_quality >= 0.5:
            bid = self.suit_bid(3, suit, '1032')
        else:
            bid = Pass('1033')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_six_six_bid(self) -> Bid:
        """Return bid for  weak hand with two six card suits."""
        if self.hcp < 6:
            bid = Pass('1034')
        elif self.spades == 6:
            bid = self.spade_bid(2, '1036')
        elif self.hearts == 6:
            bid = self.heart_bid(2, '1036')
        elif self.diamonds == 6:
            bid = self.diamond_bid(2, '1037')
        else:
            bid = Pass('1038')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _weak_seven_six_bid(self) -> Bid:
        bid = self.suit_bid(3, self.longest_suit, '1039')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _five_card_major_with_seven_suit_points(self) -> Bid:
        """Returns True if hand has 5 card major with 7 points in that suit."""
        for suit in (self.spade_suit, self.heart_suit):
            if (self.cards_in_suit(suit) == 5 and
                    self.suit_points(suit) >= 7):
                return True
        return False



""" Bid for Game
    Acol OverCallersBid module
"""

import inspect
from .bidding import Pass, Double, Call
from .hand import Hand

inspection = inspect.currentframe


class OverCallersThirdBid(Hand):
    """BfG OverCallersRebid class."""
    def __init__(self, hand_cards, board):
        super(OverCallersThirdBid, self).__init__(hand_cards, board)
        self.trace = 0

    def suggested_bid(self):
        """Direct control to relevant method and return a Bid object."""
        my_first_bid = Call(self.bid_history[-8])
        my_second_bid = Call(self.bid_history[-4])
        advancers_first_bid = Call(self.bid_history[-6])
        advancers_last_bid = Call(self.bid_history[-2])
        if (my_first_bid.is_pass and my_second_bid._is_pass and advancers_last_bid.is_pass):
            bid = Pass('2835')
        elif self._advancer_supports_major_after_stayman():
            if self.hcp >= 16:
                bid = self.next_level_bid(self.bid_two.denomination, '2801')
            else:
                bid = Pass('2802')
        elif self._can_show_second_major_after_stayman():
            bid = self.next_level_bid(self.spade_suit, '2803')
        elif self._has_thirteen_points_advancer_supports_second_suit():
            bid = self.next_level_bid(self.bid_two.denomination, '2804')
        elif self._has_fifteen_points_advancer_bids_two_suits():
            bid = self._fifteen_points_advancer_bids_two_suits()
        elif self._is_very_strong_and_five_four():
            if self.second_suit not in self.opponents_suits:
                suit = self.second_suit
            else:
                suit = self.longest_suit
            bid = self.next_level_bid(suit, '2805')
        elif self._is_strong_and_advancer_bids():
            bid = self._strong_hand()
        elif self._three_suits_have_been_bid():
            bid = self._three_suits_bid()
        elif self._has_two_card_support_for_advancers_major():
            bid = self.next_level_bid(self.advancer_suit_one, '2806')
        elif self._has_six_card_suit_advancer_repeats_suit():
            bid = self.next_level_bid(self.longest_suit, '2807')
        elif self._is_strong_advancer_bids_nt_after_stayman():
            bid = self.nt_bid(3, '2808')
        elif advancers_last_bid.is_game:
            bid = Pass('2826')
        elif (advancers_last_bid.denomination.name == self.longest_suit.name and
              self.bid_one.denomination != self.bid_two.denomination):
            bid = Pass('2842')
        elif advancers_last_bid.denomination.name == self.longest_suit.name:
            bid = Pass('2828')
        elif (my_first_bid.name == '1NT' and advancers_last_bid.name == '2NT' and
                self.hcp <= 15):
            bid = Pass('2829')
        elif (my_first_bid.name == '1NT' and advancers_last_bid.name == '2NT' and
                self.hcp <= 15):
            bid = Pass('2829')
        elif (not advancers_last_bid.is_pass and (advancers_last_bid.denomination == my_first_bid.denomination or
                    advancers_last_bid.denomination == my_second_bid.denomination)):
            bid = Pass('2830')
        elif (not advancers_last_bid.is_pass and advancers_first_bid.denomination == advancers_last_bid.denomination and
                self.suit_length(advancers_first_bid.denomination) <= 2):
            bid = Pass('2832')
        elif (not advancers_last_bid.is_pass and
              advancers_first_bid.denomination == advancers_last_bid.denomination and
                self.hcp < 16):
            bid = Pass('2833')
        elif (self.partner_bid_one.is_suit_call and self.advancer_bid_two.is_suit_call and
              self.advancer_suit_one != self.advancer_bid_two.denomination and
              self.suit_holding[self.advancer_suit_one] >= self.suit_holding[self.advancer_bid_two.denomination] + 1 and
              self.next_level(self.advancer_suit_one) < self.next_level(self.advancer_bid_two.denomination) + 1):
            bid = self.next_level_bid(self.advancer_suit_one, '2834')
        elif advancers_last_bid.is_pass:
            bid = Pass('2831')
        elif my_first_bid.is_double and my_second_bid.is_double and advancers_first_bid.is_pass:
            bid = Pass('2836')
        elif self.nt_level >= 3 and self.hcp <= 13:
            bid = Pass('2837')
        elif self.partner_last_bid.name == '2NT' and self.shape[1] <= 3:
            bid = Pass('2844')
        elif (self.advancer_bid_two.is_suit_call and
                self.suit_holding[self.advancer_bid_two.denomination] <= 3):
            bid = Pass('2845')
        elif not self.advancer_bid_two.is_pass:
            bid = Pass('2848')
        elif (self.advancer_bid_one.is_pass and
              self.advancer_bid_two.is_suit_call and
              self.suit_holding[self.advancer_suit_two] <= 3):
            bid = Pass('2850')
        else:
            bid = Pass('2809')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _strong_hand(self):
        """Return bid with strong hand."""
        if self.bid_one.is_double and self.bid_two.is_double:
            bid = Pass('2810')
        elif self._support_for_advancer():
            bid = self.next_level_bid(self.advancer_suit_one, '2811')
        elif self.is_balanced and self.nt_level <= 3:
            bid = self.next_nt_bid('2812')
        elif self.five_five and self.second_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.second_suit, '2813')
        elif self.five_four and self.second_suit not in self.opponents_suits:
            bid = self.next_level_bid(self.second_suit, '2814')
        elif (self.advancer_suit_one != self.partner_last_bid.denomination and
                self.suit_length(self.second_suit) >= 5 and
                self.second_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.second_suit, '2827')
        elif (self.partner_penultimate_bid.name == '1NT' and
              self.partner_last_bid.name == '2NT' and
              self.five_five and
              self.second_suit in self.opponents_suits and
              self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '2841')
        elif (self.advancer_suit_one != self.partner_last_bid.denomination and
                self.longest_suit not in self.opponents_suits):
            bid = self.next_level_bid(self.longest_suit, '2815')
        elif (not self.partner_bid_one.is_pass and
                self.suit_length(self.advancer_suit_one) >= 3):
            bid = self.next_level_bid(self.advancer_suit_one, '2816')
        elif self.advancer_bid_two.name == '2NT' and self.hcp <= 17:
            bid = Pass('2840')
        elif self.nt_level >= 4:
            bid = Pass('2846')
        elif self.advancer_suit_two and self.suit_holding[self.advancer_suit_two] <= 1:
            bid = Pass('2847')
        elif 1:
            bid = Pass('2849')
        else:
            bid = Pass('2817')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _three_suits_bid(self):
        """Return bid after 3 suits have been bid."""
        suit_to_bid = self._select_suit_after_three_suits_bid()
        if self._suit_is_minor(suit_to_bid):
            bid = self.next_nt_bid('2818')
        elif suit_to_bid not in self.opponents_suits:
            bid = self.next_level_bid(suit_to_bid, '2821')
        else:
            bid = Pass('2820')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    def _fifteen_points_advancer_bids_two_suits(self):
        """Return bid if strong and advancer bids two suits."""
        if self.nt_level >= 6 and self.hcp >= 16 and self.is_semi_balanced and self.double_allowed():
            bid = Double('2838')
        elif self.bidding_above_game and self.double_allowed():
            bid = Double('2825')
        elif self._advancer_has_passed_after_double():
            bid = Pass('2822')
        elif (self.advancer_bid_one.is_suit_call and
              self.advancer_suit_one != self.advancer_bid_two.denomination and
              self.suit_length(self.advancer_suit_one) >= 3
              and self.hcp >= 20):
            bid = self.next_level_bid(self.advancer_suit_one,'2839')
        elif (self.advancer_bid_one.is_suit_call and
              self.advancer_suit_one != self.advancer_bid_two.denomination and
              self.suit_length(self.advancer_bid_two.denomination) >= 4
               and self.hcp >= 20):
            # allowed duplicate comment x_refs
            bid = self.next_level_bid(self.advancer_suit_one,'2839')
        elif (self.suit_length(self.opener_suit_one) >= 3 and
                self.suit_points(self.opener_suit_one) >= 5):
            bid = Pass('2823')
        elif (self.advancer_suit_one != self.partner_bid_two.denomination and
              self.hcp >= 15):
            suit = self._suit_preference()
            bid = self.next_level_bid(suit, '2843')
        else:
            bid = Pass('2824')
        self.tracer(__name__, inspection(), bid, self.trace)
        return bid

    # Various utility functions

    def _suit_preference(self):
        """Return preference (if any) for partner's suits."""
        suit = None
        suit_one = None
        suit_two = None
        if (self.partner_bid_one.is_suit_call):
            suit_one = self.partner_bid_one.denomination
        if self.partner_last_bid.denomination.is_suit:
            suit_two = self.partner_last_bid.denomination

        if suit_one and suit_two:
            suit = suit_two
            if self.suit_length(suit_one) >= self.suit_length(suit_two) - 1:
                suit = suit_one
        else:

            if suit_one and self.suit_length(suit_one) >= 3:
                suit = suit_one
            if suit_two and self.suit_length(suit_two) >= 4:
                return suit_two
        if not suit:
            return self.no_trumps
        return suit

    def _select_suit_after_three_suits_bid(self):
        if self.suit_length(self.partner_last_bid.denomination) >= 4:
            return self.partner_last_bid.denomination
        elif self.five_five:
            suit_to_bid = self.second_suit
            if suit_to_bid in self.opponents_suits:
                return self.longest_suit
            return self.second_suit
        return  self.longest_suit

    # Various boolean functions

    def _advancer_supports_major_after_stayman(self):
        """Return True if advancer supports major after Stayman."""
        return (self.bid_one.name == '1NT' and
                self.partner_bid_one.name == '2C' and
                self.partner_last_bid.denomination == self.bid_two.denomination and
                self.partner_last_bid.level == 3)

    def _can_show_second_major_after_stayman(self):
        """Return True if got 4 spades and 4 hearts."""
        return (self.bid_one.name == '1NT' and
                self.partner_bid_one.name == '2C' and
                self.bid_two.name == '2H' and
                self.partner_last_bid.name == '2NT' and
                self.spades >= 4 and
                self.spade_suit not in self.opponents_suits)

    def _has_thirteen_points_advancer_supports_second_suit(self):
        """Return True if 13 points and advancer support second suit."""
        return (self.partner_last_bid.denomination == self.bid_two.denomination and
                self.partner_last_bid.is_suit_call and
                not self.partner_last_bid.is_game and
                self.hcp >= 13 and
                not self.singleton_honour)

    def _has_fifteen_points_advancer_bids_two_suits(self):
        """Return True if fifteen points and advancer bids two suits."""
        return (self.advancer_suit_one != self.partner_last_bid.denomination and
                not self.bid_one.is_nt and
                not self.partner_last_bid.is_nt and
                not self.partner_last_bid.is_game and
                self._suit_preference().is_suit and
                self.hcp >= 15)

    def _is_very_strong_and_five_four(self):
        """Return True if strong and 5/4."""
        return (self.hcp >= 19 and
                self.five_four_or_better and
                not self.bid_two.is_nt and
                not self.bidding_above_game and
                self.longest_suit not in self.opponents_suits)

    def _is_strong_and_advancer_bids(self):
        """Return True if strong and advancer bids."""
        return (self.hcp >= 16 and
                not self.partner_last_bid.is_pass and
                not self.partner_last_bid.is_game)

    def _three_suits_have_been_bid(self):
        """Return True if 3 suits bid."""
        my_suits = [self.bid_one.denomination, self.bid_two.denomination]
        return (self.advancer_suit_one != self.partner_last_bid.denomination and
                self.partner_bid_one.is_suit_call and
                self.partner_last_bid.level == 3 and
                self.bid_one.denomination != self.partner_last_bid.denomination and
                not self.partner_last_bid.is_game and
                self.partner_last_bid.denomination not in my_suits)

    def _advancer_has_passed_after_double(self):
        """Return True if advancer passes after double."""
        return (self.bid_one.is_double and
                (self.partner_bid_one.is_pass or
                 self.partner_last_bid.is_pass))

    def _has_two_card_support_for_advancers_major(self):
        """Return True if 2 card support for advancers major."""
        return (self.advancer_suit_one == self.partner_last_bid.denomination and
                self.partner_bid_one.is_value_call and
                self.advancer_suit_one.is_major and
                self.suit_length(self.advancer_suit_one) >= 2 and
                self.hcp >= 15 and
                not self.partner_last_bid.is_game)
        return result

    def _has_six_card_suit_advancer_repeats_suit(self):
        """Return True if 6 card suit and advancer repeats their suit."""
        return (self.advancer_suit_one == self.partner_last_bid.denomination and
                self.partner_bid_one.is_value_call and
                self.shape[0] >= 6 and
                not self.partner_last_bid.is_game and
                self.longest_suit not in self.opponents_suits and
                self.hcp >= 12)

    def _is_strong_advancer_bids_nt_after_stayman(self):
        """Return True if strong and advancer bids NT after stayman."""
        return (self.bid_one.name == '1NT' and
                self.partner_bid_one.name == '2C' and
                self.partner_last_bid.name == '2NT' and
                self.hcp >= 17 and
                self.nt_level <= 3)

    def _suit_is_minor(self, suit_to_bid):
        """Return True if suit is minor."""
        return (self.nt_level <= 3 and
                (suit_to_bid.is_minor or
                suit_to_bid in self.opponents_suits))

    def _support_for_advancer(self):
        """Return True if xxx."""
        if not self.advancer_bid_two:
           return False
        return (self.advancer_suit_one == self.advancer_bid_two.denomination and
                self.suit_length(self.advancer_suit_one) >= 3)



""" Bid for Game
    Bridge objects
"""

from bridgeobjects import Card, SUITS, Hand, Suit
from bridgeobjects import Call
from .comments import comment_id, comment_html, strategy_html
# from .strategy_xref import StrategyXref


class Bid(Call):
    """Return BfG Bid class."""
    def __init__(self, name: str = '',
                 call_id: str = '0000',
                 use_shortage_points:
                 bool = False,
                 *args,
                 **kwargs) -> None:
        super().__init__(name, *args, **kwargs)
        self.use_shortage_points: bool = use_shortage_points
        self.rtf_comment: str = ''
        self.rtf_strategy: str = ''
        # if call_id not in CommentXref.comments:
        #     call_id = '0000'
        self.call_id: str = call_id

        # the comments dict contains a gettext reference to the
        # comment/strategy text.
        self.comment_id: str = ''
        self.strategy_id: str = ''
        self.comment_html: str = ''
        self.strategy_html: str = ''
        self.comment_rst: str = ''
        self.strategy_rst: str = ''

    def __repr__(self) -> str:
        """Return a __repr__ display value of object."""
        return f'bid: {self._name}, {self.call_id}'

    def get_comments(self) -> None:
        """Get the bid comments from the x_refs."""
        self.comment_id = comment_id(self.call_id)
        self.comment_html = comment_html(self.call_id)
        self.strategy_html = strategy_html(self.call_id)

        self.comment_rst = self._convert_html_to_rst(self.comment_html)
        self.strategy_rst = self._convert_html_to_rst(self.strategy_html)

    @staticmethod
    def _convert_html_to_rst(html: str) -> str:
        """Return and html_string as rst."""
        rst = html.replace('<br>', '\n')
        rst = rst.replace('<p> ', '<p>')
        rst = rst.replace('<p>', '\n\n')
        rst = rst.replace('<span style="color:red">', '**')
        rst = rst.replace('<span style="color:green">', '**')
        rst = rst.replace('</span>', '**')
        return rst


class Pass(Bid):
    """Class definition for PASS bid."""
    def __init__(
            self, call_id: str = '0000',
            use_shortage_points: bool = False,
            *args, **kwargs):
        super(Pass, self).__init__(
            name='P',
            call_id=call_id,
            use_shortage_points=use_shortage_points,
            *args, **kwargs)
        pass


class Double(Bid):
    """Class definition for DOUBLE bid."""
    def __init__(
            self,
            call_id: str = '0000',
            use_shortage_points: bool = False,
            *args, **kwargs):
        super(Double, self).__init__(
            name='D',
            call_id=call_id,
            use_shortage_points=use_shortage_points,
            *args, **kwargs)
        pass


class HandSuit(object):
    """Instantiate BfG HandSuit class."""
    # SUITS = {
    #     'NT': -1,
    #     'S': 0,
    #     'H': 1,
    #     'D': 2,
    #     'C': 3,
    #     -1: 'NT',
    #     0: 'S',
    #     1: 'H',
    #     2: 'D',
    #     3: 'C'
    # }

    def __init__(self, suit: Suit | None = None,
                 hand: Hand | None = None,
                 bid_history: list[str] = None):
        """Initialise class."""
        self.suit = suit
        self.name = suit.name
        self.cards = []
        self._card_list = []
        if hand:
            self.cards = hand.cards
            self._card_list = [card for card in hand.cards
                               if card.suit == suit]
        if not bid_history:
            bid_history = []
        self.bid_history = bid_history
        self.honours = 0
        self.touching_honours = 0
        self.honour_points = 0
        self.ace_as_single_honour = False
        self.highest_touching_honour = -1
        if self._card_list:
            self.honours = self._honours()
            self.touching_honours = self._touching_honours()
            self.highest_touching_honour = self._highest_touching_honour()
            self.ace_as_single_honour = self._ace_as_single_honour()
            self.length = len(self._card_list)
            self.honour_points = self._honour_points()
        if bid_history:
            (self.partners_bid, self.opponents_bid) = self._get_bids()

    def __repr__(self) -> str:
        """Return a __repr__ display value of object."""
        return 'HandSuit: %s' % self.name

    def suit_quality(self) -> int:
        """Return the suit quality."""
        quality = 0
        if Card('A', self.suit.name) in self._card_list:
            quality += 1
        if Card('K', self.suit.name) in self._card_list:
            quality += 1
        if Card('Q', self.suit.name) in self._card_list:
            quality += 0.5
        if Card('J', self.suit.name) in self._card_list:
            quality += 0.5
        return quality

    def has_control(self, hand: Hand) -> bool:
        """Return True if hand has control of that suit."""
        suit = self.suit
        if Card('A', suit.name) in self.cards:
            return True
        elif (Card('K', suit.name) in self.cards and
              hand.suit_holding[suit] >= 2):
            return True
        elif (Card('Q', suit.name) in self.cards and
              hand.suit_holding[suit] > 3):
            return True
        elif (Card('J', suit.name) in self.cards and
              hand.suit_holding[suit] >= 4):
            return True
        return False

    def _touching_honours(self) -> int:
        touching_pairs = ['AK', 'KQ', 'QJ', 'JT']
        touching_honours = 1
        touching_honours_hold = 1
        for index, card in enumerate(self._card_list[1:]):
            if self._card_list[index].rank + card.rank in touching_pairs:
                touching_honours += 1
            else:
                if touching_honours_hold < touching_honours:
                    touching_honours_hold = touching_honours
                touching_honours = 1
        touching_honours = max(touching_honours, touching_honours_hold)
        return touching_honours

    def _highest_touching_honour(self) -> Card | None:
        highest_touching_honour = None
        if self.cards and self.touching_honours > 1:
            highest_touching_honour = self.cards[0]
            for card in self.cards:
                if card.value == 13 - highest_touching_honour.value:
                    highest_touching_honour = card
                    break
        return highest_touching_honour

    def _ace_as_single_honour(self) -> bool:
        result = False
        for card in self._card_list:
            if card.name == 'A':
                result = True
            elif card.name in ['TJQK']:
                result = False
                break
        return result

    def _honours(self) -> int:
        """Return the number of honours in hand_suit including tens."""
        honours = 0
        for card in self._card_list:
            if card.value >= 9:
                honours += 1
        return honours

    def _honour_points(self) -> int:
        value = 0
        for index, card in enumerate(self._card_list[:5]):
            if card:
                value += 4 - index
        return value

    def _get_bids(self) -> tuple[bool]:
        (partners_bid, opponents_bid) = (False, False)
        bid_history = list(reversed(self.bid_history[:-3]))
        offset = 0
        contract_suit = self._suit_from_bid(bid_history[0])
        for index, bid in enumerate(bid_history[0::2]):
            suit = self._suit_from_bid(bid)
            if suit == self.suit:
                opponents_bid = True
            if contract_suit == self._suit_from_bid(bid):
                offset = index % 2

        for bid in bid_history[offset*2+1::4]:
            suit = self._suit_from_bid(bid)
            if suit == self.suit:
                partners_bid = True

        return (partners_bid, opponents_bid)

    @staticmethod
    def _suit_from_bid(bid: str) -> Suit | int | None:
        """Return suit from a bid."""
        suit = None
        if len(bid) == 2:
            suit = SUITS[bid[1]].name
        elif len(bid) == 2:
            suit = -1
        return suit



"""Expose the classes in the API."""
from ._version import __version__
# from icecream import ic, install
# ic.configureOutput(includeContext=True)
# install()


from .hand import Hand
from .comments import comments, strategies, comment_xrefs, convert_text_to_html
from .strategy_xref import StrategyXref, strategy_descriptions
from .bidding import Bid, Pass, Double
from .player import Player
from .utils import get_role

VERSION = __version__
